# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import contextlib
import os
import re
from tempfile import NamedTemporaryFile
import threading

import requests

from ..common import ContentLocation
from ..locale import gettext_lazy as _
from ..url import PrivacyMode
from .base import AttachBase


class AttachHTTP(AttachBase):
    """A wrapper for HTTP based attachment sources."""

    # The default descriptive name associated with the service
    service_name = _("Web Based")

    # The default protocol
    protocol = "http"

    # The default secure protocol
    secure_protocol = "https"

    # The number of bytes in memory to read from the remote source at a time
    chunk_size = 8192

    # Web based requests are remote/external to our current location
    location = ContentLocation.HOSTED

    # thread safe loading
    _lock = threading.Lock()

    def __init__(self, headers=None, **kwargs):
        """Initialize HTTP Object.

        headers can be a dictionary of key/value pairs that you want to
        additionally include as part of the server headers to post with
        """
        super().__init__(**kwargs)

        self.schema = "https" if self.secure else "http"

        self.fullpath = kwargs.get("fullpath")
        if not isinstance(self.fullpath, str):
            self.fullpath = "/"

        self.headers = {}
        if headers:
            # Store our extra headers
            self.headers.update(headers)

        # Where our content is written to upon a call to download.
        self._temp_file = None

        # Our Query String Dictionary; we use this to track arguments
        # specified that aren't otherwise part of this class
        self.qsd = {
            k: v
            for k, v in kwargs.get("qsd", {}).items()
            if k not in self.template_args
        }

        return

    def download(self, **kwargs):
        """Perform retrieval of the configuration based on the specified
        request."""

        if self.location == ContentLocation.INACCESSIBLE:
            # our content is inaccessible
            return False

        # prepare header
        headers = {
            "User-Agent": self.app_id,
        }

        # Apply any/all header over-rides defined
        headers.update(self.headers)

        auth = None
        if self.user:
            auth = (self.user, self.password)

        url = f"{self.schema}://{self.host}"
        if isinstance(self.port, int):
            url += f":{self.port}"

        url += self.fullpath

        # Where our request object will temporarily live.
        r = None

        # Always call throttle before any remote server i/o is made
        self.throttle()

        with self._lock:
            if self.exists(retrieve_if_missing=False):
                # Due to locking; it's possible a concurrent thread already
                # handled the retrieval in which case we can safely move on
                self.logger.trace(
                    "HTTP Attachment %s already retrieved",
                    self._temp_file.name,
                )
                return True

            # Ensure any existing content set has been invalidated
            self.invalidate()

            self.logger.debug(
                "HTTP Attachment Fetch URL:"
                f" {url} (cert_verify={self.verify_certificate!r})"
            )

            try:
                # Make our request
                with requests.get(
                    url,
                    headers=headers,
                    auth=auth,
                    params=self.qsd,
                    verify=self.verify_certificate,
                    timeout=self.request_timeout,
                    stream=True,
                ) as r:

                    # Handle Errors
                    r.raise_for_status()

                    # Get our file-size (if known)
                    try:
                        file_size = int(r.headers.get("Content-Length", "0"))
                    except (TypeError, ValueError):
                        # Handle edge case where Content-Length is a bad value
                        file_size = 0

                    # Perform a little Q/A on file limitations and restrictions
                    if (
                        self.max_file_size > 0
                        and file_size > self.max_file_size
                    ):

                        # The content retrieved is to large
                        self.logger.error(
                            "HTTP response exceeds allowable maximum file"
                            f" length ({int(self.max_file_size / 1024)}KB):"
                            f" {self.url(privacy=True)}"
                        )

                        # Return False (signifying a failure)
                        return False

                    # Detect config format based on mime if the format isn't
                    # already enforced
                    self.detected_mimetype = r.headers.get("Content-Type")

                    d = r.headers.get("Content-Disposition", "")
                    result = re.search(
                        "filename=['\"]?(?P<name>[^'\"]+)['\"]?", d, re.I
                    )
                    if result:
                        self.detected_name = result.group("name").strip()

                    # Create a temporary file to work with; delete must be set
                    # to False or it isn't compatible with Microsoft Windows
                    # instances. In lieu of this, __del__ will clean up the
                    # file for us.
                    self._temp_file = \
                        NamedTemporaryFile(delete=False)  # noqa: SIM115

                    # Get our chunk size
                    chunk_size = self.chunk_size

                    # Track all bytes written to disk
                    bytes_written = 0

                    # If we get here, we can now safely write our content to
                    # disk
                    for chunk in r.iter_content(chunk_size=chunk_size):
                        # filter out keep-alive chunks
                        if chunk:
                            self._temp_file.write(chunk)
                            bytes_written = self._temp_file.tell()

                            # Prevent a case where Content-Length isn't
                            # provided. In this case we don't want to fetch
                            # beyond our limits
                            if self.max_file_size > 0:
                                if bytes_written > self.max_file_size:
                                    # The content retrieved is to large
                                    self.logger.error(
                                        "HTTP response exceeds allowable"
                                        " maximum file length"
                                        f" ({int(self.max_file_size / 1024)}"
                                        f"KB): {self.url(privacy=True)}"
                                    )

                                    # Invalidate any variables previously set
                                    self.invalidate()

                                    # Return False (signifying a failure)
                                    return False

                                elif (
                                    bytes_written + chunk_size
                                    > self.max_file_size
                                ):
                                    # Adjust out next read to accomodate up to
                                    # our limit +1. This will prevent us from
                                    # reading to much into our memory buffer
                                    self.max_file_size - bytes_written + 1

                    # Ensure our content is flushed to disk for post-processing
                    self._temp_file.flush()

                    # Set our minimum requirements for a successful download()
                    # call
                    self.download_path = self._temp_file.name
                    if not self.detected_name:
                        self.detected_name = os.path.basename(self.fullpath)

            except requests.RequestException as e:
                self.logger.error(
                    "A Connection error occurred retrieving HTTP "
                    f"configuration from {self.host}."
                )
                self.logger.debug(f"Socket Exception: {e!s}")

                # Invalidate any variables previously set
                self.invalidate()

                # Return False (signifying a failure)
                return False

            except OSError:
                # IOError is present for backwards compatibility with Python
                # versions older then 3.3.  >= 3.3 throw OSError now.

                # Could not open and/or write the temporary file
                self.logger.error(
                    "Could not write attachment to disk:"
                    f" {self.url(privacy=True)}"
                )

                # Invalidate any variables previously set
                self.invalidate()

                # Return False (signifying a failure)
                return False

        # Return our success
        return True

    def invalidate(self):
        """Close our temporary file."""
        if self._temp_file:
            self.logger.trace("Attachment cleanup of %s", self._temp_file.name)
            self._temp_file.close()

            with contextlib.suppress(OSError):
                # Ensure our file is removed (if it exists)
                os.unlink(self._temp_file.name)

            # Reset our temporary file to prevent from entering
            # this block again
            self._temp_file = None

        super().invalidate()

    def __del__(self):
        """Tidy memory if open."""
        self.invalidate()

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Our URL parameters
        params = self.url_parameters(privacy=privacy, *args, **kwargs)

        # Prepare our cache value
        if self.cache is not None:
            if isinstance(self.cache, bool) or not self.cache:
                cache = "yes" if self.cache else "no"
            else:
                cache = int(self.cache)

            # Set our cache value
            params["cache"] = cache

        if self._mimetype:
            # A format was enforced
            params["mime"] = self._mimetype

        if self._name:
            # A name was enforced
            params["name"] = self._name

        # Append our headers into our parameters
        params.update({f"+{k}": v for k, v in self.headers.items()})

        # Apply any remaining entries to our URL
        params.update(self.qsd)

        # Determine Authentication
        auth = ""
        if self.user and self.password:
            auth = "{user}:{password}@".format(
                user=self.quote(self.user, safe=""),
                password=self.pprint(
                    self.password, privacy, mode=PrivacyMode.Secret, safe=""
                ),
            )
        elif self.user:
            auth = "{user}@".format(
                user=self.quote(self.user, safe=""),
            )

        default_port = 443 if self.secure else 80
        return "{schema}://{auth}{hostname}{port}{fullpath}?{params}".format(
            schema=self.secure_protocol if self.secure else self.protocol,
            auth=auth,
            hostname=self.quote(self.host, safe=""),
            port=(
                ""
                if self.port is None or self.port == default_port
                else f":{self.port}"
            ),
            fullpath=self.quote(self.fullpath, safe="/"),
            params=self.urlencode(params, safe="/"),
        )

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = AttachBase.parse_url(url, sanitize=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Add our headers that the user can potentially over-ride if they wish
        # to to our returned result set
        results["headers"] = results["qsd-"]
        results["headers"].update(results["qsd+"])

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import contextlib

import requests

from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import validate_regex
from .base import NotifyBase


# Priorities
class ProwlPriority:
    LOW = -2
    MODERATE = -1
    NORMAL = 0
    HIGH = 1
    EMERGENCY = 2


PROWL_PRIORITIES = {
    # Note: This also acts as a reverse lookup mapping
    ProwlPriority.LOW: "low",
    ProwlPriority.MODERATE: "moderate",
    ProwlPriority.NORMAL: "normal",
    ProwlPriority.HIGH: "high",
    ProwlPriority.EMERGENCY: "emergency",
}

PROWL_PRIORITY_MAP = {
    # Maps against string 'low'
    "l": ProwlPriority.LOW,
    # Maps against string 'moderate'
    "m": ProwlPriority.MODERATE,
    # Maps against string 'normal'
    "n": ProwlPriority.NORMAL,
    # Maps against string 'high'
    "h": ProwlPriority.HIGH,
    # Maps against string 'emergency'
    "e": ProwlPriority.EMERGENCY,
    # Entries to additionally support (so more like Prowl's API)
    "-2": ProwlPriority.LOW,
    "-1": ProwlPriority.MODERATE,
    "0": ProwlPriority.NORMAL,
    "1": ProwlPriority.HIGH,
    "2": ProwlPriority.EMERGENCY,
}

# Provide some known codes Prowl uses and what they translate to:
PROWL_HTTP_ERROR_MAP = {
    406: "IP address has exceeded API limit",
    409: "Request not aproved.",
}


class NotifyProwl(NotifyBase):
    """A wrapper for Prowl Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Prowl"

    # The services URL
    service_url = "https://www.prowlapp.com/"

    # The default secure protocol
    secure_protocol = "prowl"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_prowl"

    # Prowl uses the http protocol with JSON requests
    notify_url = "https://api.prowlapp.com/publicapi/add"

    # Disable throttle rate for Prowl requests since they are normally
    # local anyway
    request_rate_per_sec = 0

    # The maximum allowable characters allowed in the body per message
    body_maxlen = 10000

    # Defines the maximum allowable characters in the title
    title_maxlen = 1024

    # Define object templates
    templates = (
        "{schema}://{apikey}",
        "{schema}://{apikey}/{providerkey}",
    )

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "apikey": {
                "name": _("API Key"),
                "type": "string",
                "private": True,
                "required": True,
                "regex": (r"^[A-Za-z0-9]{40}$", "i"),
            },
            "providerkey": {
                "name": _("Provider Key"),
                "type": "string",
                "private": True,
                "regex": (r"^[A-Za-z0-9]{40}$", "i"),
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "priority": {
                "name": _("Priority"),
                "type": "choice:int",
                "values": PROWL_PRIORITIES,
                "default": ProwlPriority.NORMAL,
            },
        },
    )

    def __init__(self, apikey, providerkey=None, priority=None, **kwargs):
        """Initialize Prowl Object."""
        super().__init__(**kwargs)

        # The Priority of the message
        self.priority = (
            NotifyProwl.template_args["priority"]["default"]
            if not priority
            else next(
                (
                    v
                    for k, v in PROWL_PRIORITY_MAP.items()
                    if str(priority).lower().startswith(k)
                ),
                NotifyProwl.template_args["priority"]["default"],
            )
        )

        # API Key (associated with project)
        self.apikey = validate_regex(
            apikey, *self.template_tokens["apikey"]["regex"]
        )
        if not self.apikey:
            msg = f"An invalid Prowl API Key ({apikey}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        # Store the provider key (if specified)
        if providerkey:
            self.providerkey = validate_regex(
                providerkey, *self.template_tokens["providerkey"]["regex"]
            )
            if not self.providerkey:
                msg = (
                    "An invalid Prowl Provider Key "
                    f"({providerkey}) was specified."
                )
                self.logger.warning(msg)
                raise TypeError(msg)

        else:
            # No provider key was set
            self.providerkey = None

        return

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform Prowl Notification."""

        headers = {
            "User-Agent": self.app_id,
            "Content-type": "application/x-www-form-urlencoded",
        }

        # prepare JSON Object
        payload = {
            "apikey": self.apikey,
            "application": self.app_id,
            "event": title,
            "description": body,
            "priority": self.priority,
        }

        if self.providerkey:
            payload["providerkey"] = self.providerkey

        self.logger.debug(
            "Prowl POST URL:"
            f" {self.notify_url} (cert_verify={self.verify_certificate!r})"
        )
        self.logger.debug(f"Prowl Payload: {payload!s}")

        # Always call throttle before any remote server i/o is made
        self.throttle()

        try:
            r = requests.post(
                self.notify_url,
                data=payload,
                headers=headers,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )
            if r.status_code != requests.codes.ok:
                # We had a problem
                status_str = NotifyBase.http_response_code_lookup(
                    r.status_code, PROWL_HTTP_ERROR_MAP
                )

                self.logger.warning(
                    "Failed to send Prowl notification:{}{}error={}.".format(
                        status_str, ", " if status_str else "", r.status_code
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                # Return; we're done
                return False

            else:
                self.logger.info("Sent Prowl notification.")

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred sending Prowl notification."
            )
            self.logger.debug(f"Socket Exception: {e!s}")

            # Return; we're done
            return False

        return True

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (self.secure_protocol, self.apikey, self.providerkey)

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "priority": (
                PROWL_PRIORITIES[self.template_args["priority"]["default"]]
                if self.priority not in PROWL_PRIORITIES
                else PROWL_PRIORITIES[self.priority]
            ),
        }

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        return "{schema}://{apikey}/{providerkey}/?{params}".format(
            schema=self.secure_protocol,
            apikey=self.pprint(self.apikey, privacy, safe=""),
            providerkey=self.pprint(self.providerkey, privacy, safe=""),
            params=NotifyProwl.urlencode(params),
        )

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Set the API Key
        results["apikey"] = NotifyProwl.unquote(results["host"])

        # Optionally try to find the provider key
        with contextlib.suppress(IndexError):
            results["providerkey"] = NotifyProwl.split_path(
                results["fullpath"]
            )[0]

        # Set our priority
        if "priority" in results["qsd"] and len(results["qsd"]["priority"]):
            results["priority"] = NotifyProwl.unquote(
                results["qsd"]["priority"]
            )

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from json import loads

import requests

from .. import exception
from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import parse_list, validate_regex
from .base import NotifyBase


class PushSaferSound:
    """Defines all of the supported PushSafe sounds."""

    # Silent
    SILENT = 0
    # Ahem (IM)
    AHEM = 1
    # Applause (Mail)
    APPLAUSE = 2
    # Arrow (Reminder)
    ARROW = 3
    # Baby (SMS)
    BABY = 4
    # Bell (Alarm)
    BELL = 5
    # Bicycle (Alarm2)
    BICYCLE = 6
    # Boing (Alarm3)
    BOING = 7
    # Buzzer (Alarm4)
    BUZZER = 8
    # Camera (Alarm5)
    CAMERA = 9
    # Car Horn (Alarm6)
    CAR_HORN = 10
    # Cash Register (Alarm7)
    CASH_REGISTER = 11
    # Chime (Alarm8)
    CHIME = 12
    # Creaky Door (Alarm9)
    CREAKY_DOOR = 13
    # Cuckoo Clock (Alarm10)
    CUCKOO_CLOCK = 14
    # Disconnect (Call)
    DISCONNECT = 15
    # Dog (Call2)
    DOG = 16
    # Doorbell (Call3)
    DOORBELL = 17
    # Fanfare (Call4)
    FANFARE = 18
    # Gun Shot (Call5)
    GUN_SHOT = 19
    # Honk (Call6)
    HONK = 20
    # Jaw Harp (Call7)
    JAW_HARP = 21
    # Morse (Call8)
    MORSE = 22
    # Electricity (Call9)
    ELECTRICITY = 23
    # Radio Tuner (Call10)
    RADIO_TURNER = 24
    # Sirens
    SIRENS = 25
    # Military Trumpets
    MILITARY_TRUMPETS = 26
    # Ufo
    UFO = 27
    # Whah Whah Whah
    LONG_WHAH = 28
    # Man Saying Goodbye
    GOODBYE = 29
    # Man Saying Hello
    HELLO = 30
    # Man Saying No
    NO = 31
    # Man Saying Ok
    OKAY = 32
    # Man Saying Ooohhhweee
    OOOHHHWEEE = 33
    # Man Saying Warning
    WARNING = 34
    # Man Saying Welcome
    WELCOME = 35
    # Man Saying Yeah
    YEAH = 36
    # Man Saying Yes
    YES = 37
    # Beep short
    BEEP1 = 38
    # Weeeee short
    WEEE = 39
    # Cut in and out short
    CUTINOUT = 40
    # Finger flicking glas short
    FLICK_GLASS = 41
    # Wa Wa Waaaa short
    SHORT_WHAH = 42
    # Laser short
    LASER = 43
    # Wind Chime short
    WIND_CHIME = 44
    # Echo short
    ECHO = 45
    # Zipper short
    ZIPPER = 46
    # HiHat short
    HIHAT = 47
    # Beep 2 short
    BEEP2 = 48
    # Beep 3 short
    BEEP3 = 49
    # Beep 4 short
    BEEP4 = 50
    # The Alarm is armed
    ALARM_ARMED = 51
    # The Alarm is disarmed
    ALARM_DISARMED = 52
    # The Backup is ready
    BACKUP_READY = 53
    # The Door is closed
    DOOR_CLOSED = 54
    # The Door is opend
    DOOR_OPENED = 55
    # The Window is closed
    WINDOW_CLOSED = 56
    # The Window is open
    WINDOW_OPEN = 57
    # The Light is off
    LIGHT_ON = 58
    # The Light is on
    LIGHT_OFF = 59
    # The Doorbell rings
    DOORBELL_RANG = 60


PUSHSAFER_SOUND_MAP = {
    # Device Default,
    "silent": PushSaferSound.SILENT,
    "ahem": PushSaferSound.AHEM,
    "applause": PushSaferSound.APPLAUSE,
    "arrow": PushSaferSound.ARROW,
    "baby": PushSaferSound.BABY,
    "bell": PushSaferSound.BELL,
    "bicycle": PushSaferSound.BICYCLE,
    "bike": PushSaferSound.BICYCLE,
    "boing": PushSaferSound.BOING,
    "buzzer": PushSaferSound.BUZZER,
    "camera": PushSaferSound.CAMERA,
    "carhorn": PushSaferSound.CAR_HORN,
    "horn": PushSaferSound.CAR_HORN,
    "cashregister": PushSaferSound.CASH_REGISTER,
    "chime": PushSaferSound.CHIME,
    "creakydoor": PushSaferSound.CREAKY_DOOR,
    "cuckooclock": PushSaferSound.CUCKOO_CLOCK,
    "cuckoo": PushSaferSound.CUCKOO_CLOCK,
    "disconnect": PushSaferSound.DISCONNECT,
    "dog": PushSaferSound.DOG,
    "doorbell": PushSaferSound.DOORBELL,
    "fanfare": PushSaferSound.FANFARE,
    "gunshot": PushSaferSound.GUN_SHOT,
    "honk": PushSaferSound.HONK,
    "jawharp": PushSaferSound.JAW_HARP,
    "morse": PushSaferSound.MORSE,
    "electric": PushSaferSound.ELECTRICITY,
    "radiotuner": PushSaferSound.RADIO_TURNER,
    "sirens": PushSaferSound.SIRENS,
    "militarytrumpets": PushSaferSound.MILITARY_TRUMPETS,
    "military": PushSaferSound.MILITARY_TRUMPETS,
    "trumpets": PushSaferSound.MILITARY_TRUMPETS,
    "ufo": PushSaferSound.UFO,
    "whahwhah": PushSaferSound.LONG_WHAH,
    "whah": PushSaferSound.SHORT_WHAH,
    "goodye": PushSaferSound.GOODBYE,
    "hello": PushSaferSound.HELLO,
    "no": PushSaferSound.NO,
    "okay": PushSaferSound.OKAY,
    "ok": PushSaferSound.OKAY,
    "ooohhhweee": PushSaferSound.OOOHHHWEEE,
    "warn": PushSaferSound.WARNING,
    "warning": PushSaferSound.WARNING,
    "welcome": PushSaferSound.WELCOME,
    "yeah": PushSaferSound.YEAH,
    "yes": PushSaferSound.YES,
    "beep": PushSaferSound.BEEP1,
    "beep1": PushSaferSound.BEEP1,
    "weee": PushSaferSound.WEEE,
    "wee": PushSaferSound.WEEE,
    "cutinout": PushSaferSound.CUTINOUT,
    "flickglass": PushSaferSound.FLICK_GLASS,
    "laser": PushSaferSound.LASER,
    "windchime": PushSaferSound.WIND_CHIME,
    "echo": PushSaferSound.ECHO,
    "zipper": PushSaferSound.ZIPPER,
    "hihat": PushSaferSound.HIHAT,
    "beep2": PushSaferSound.BEEP2,
    "beep3": PushSaferSound.BEEP3,
    "beep4": PushSaferSound.BEEP4,
    "alarmarmed": PushSaferSound.ALARM_ARMED,
    "armed": PushSaferSound.ALARM_ARMED,
    "alarmdisarmed": PushSaferSound.ALARM_DISARMED,
    "disarmed": PushSaferSound.ALARM_DISARMED,
    "backupready": PushSaferSound.BACKUP_READY,
    "dooropen": PushSaferSound.DOOR_OPENED,
    "dopen": PushSaferSound.DOOR_OPENED,
    "doorclosed": PushSaferSound.DOOR_CLOSED,
    "dclosed": PushSaferSound.DOOR_CLOSED,
    "windowopen": PushSaferSound.WINDOW_OPEN,
    "wopen": PushSaferSound.WINDOW_OPEN,
    "windowclosed": PushSaferSound.WINDOW_CLOSED,
    "wclosed": PushSaferSound.WINDOW_CLOSED,
    "lighton": PushSaferSound.LIGHT_ON,
    "lon": PushSaferSound.LIGHT_ON,
    "lightoff": PushSaferSound.LIGHT_OFF,
    "loff": PushSaferSound.LIGHT_OFF,
    "doorbellrang": PushSaferSound.DOORBELL_RANG,
}


# Priorities
class PushSaferPriority:
    LOW = -2
    MODERATE = -1
    NORMAL = 0
    HIGH = 1
    EMERGENCY = 2


PUSHSAFER_PRIORITIES = (
    PushSaferPriority.LOW,
    PushSaferPriority.MODERATE,
    PushSaferPriority.NORMAL,
    PushSaferPriority.HIGH,
    PushSaferPriority.EMERGENCY,
)

PUSHSAFER_PRIORITY_MAP = {
    # short for 'low'
    "low": PushSaferPriority.LOW,
    # short for 'medium'
    "medium": PushSaferPriority.MODERATE,
    # short for 'normal'
    "normal": PushSaferPriority.NORMAL,
    # short for 'high'
    "high": PushSaferPriority.HIGH,
    # short for 'emergency'
    "emergency": PushSaferPriority.EMERGENCY,
}

# Identify the priority ou want to designate as the fall back
DEFAULT_PRIORITY = "normal"


# Vibrations
class PushSaferVibration:
    """Defines the acceptable vibration settings for notification."""

    # x1
    LOW = 1
    # x2
    NORMAL = 2
    # x3
    HIGH = 3


# Identify all of the vibrations in one place
PUSHSAFER_VIBRATIONS = (
    PushSaferVibration.LOW,
    PushSaferVibration.NORMAL,
    PushSaferVibration.HIGH,
)

# At this time, the following pictures can be attached to each notification
# at one time. When more are supported, just add their argument below
PICTURE_PARAMETER = (
    "p",
    "p2",
    "p3",
)


# Flag used as a placeholder to sending to all devices
PUSHSAFER_SEND_TO_ALL = "a"


class NotifyPushSafer(NotifyBase):
    """A wrapper for PushSafer Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Pushsafer"

    # The services URL
    service_url = "https://www.pushsafer.com/"

    # The default insecure protocol
    protocol = "psafer"

    # The default secure protocol
    secure_protocol = "psafers"

    # Support attachments
    attachment_support = True

    # Number of requests to a allow per second
    request_rate_per_sec = 1.2

    # The icon ID of 25 looks like a megaphone
    default_pushsafer_icon = 25

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_pushsafer"

    # Defines the hostname to post content to; since this service supports
    # both insecure and secure methods, we set the {schema} just before we
    # post the message upstream.
    notify_url = "{schema}://www.pushsafer.com/api"

    # Define object templates
    templates = (
        "{schema}://{privatekey}",
        "{schema}://{privatekey}/{targets}",
    )

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "privatekey": {
                "name": _("Private Key"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "target_device": {
                "name": _("Target Device"),
                "type": "string",
                "map_to": "targets",
            },
            "target_email": {
                "name": _("Target Email"),
                "type": "string",
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "priority": {
                "name": _("Priority"),
                "type": "choice:int",
                "values": PUSHSAFER_PRIORITIES,
            },
            "sound": {
                "name": _("Sound"),
                "type": "choice:string",
                "values": PUSHSAFER_SOUND_MAP,
            },
            "vibration": {
                "name": _("Vibration"),
                "type": "choice:int",
                "values": PUSHSAFER_VIBRATIONS,
            },
            "to": {
                "alias_of": "targets",
            },
        },
    )

    def __init__(
        self,
        privatekey,
        targets=None,
        priority=None,
        sound=None,
        vibration=None,
        **kwargs,
    ):
        """Initialize PushSafer Object."""
        super().__init__(**kwargs)

        #
        # Priority
        #
        try:
            # Acquire our priority if we can:
            #  - We accept both the integer form as well as a string
            #    representation
            self.priority = int(priority)

        except TypeError:
            # NoneType means use Default; this is an okay exception
            self.priority = None

        except ValueError:
            # Input is a string; attempt to get the lookup from our
            # priority mapping
            priority = priority.lower().strip()

            # This little bit of black magic allows us to match against
            # low, lo, l (for low);
            # normal, norma, norm, nor, no, n (for normal)
            # ... etc
            match = (
                next(
                    (
                        key
                        for key in PUSHSAFER_PRIORITY_MAP
                        if key.startswith(priority)
                    ),
                    None,
                )
                if priority
                else None
            )

            # Now test to see if we got a match
            if not match:
                msg = (
                    "An invalid PushSafer priority "
                    f"({priority}) was specified."
                )
                self.logger.warning(msg)
                raise TypeError(msg) from None

            # store our successfully looked up priority
            self.priority = PUSHSAFER_PRIORITY_MAP[match]

        if (
            self.priority is not None
            and self.priority not in PUSHSAFER_PRIORITY_MAP.values()
        ):
            msg = f"An invalid PushSafer priority ({priority}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg) from None

        #
        # Sound
        #
        try:
            # Acquire our sound if we can:
            #  - We accept both the integer form as well as a string
            #    representation
            self.sound = int(sound)

        except TypeError:
            # NoneType means use Default; this is an okay exception
            self.sound = None

        except ValueError:
            # Input is a string; attempt to get the lookup from our
            # sound mapping
            sound = sound.lower().strip()

            # This little bit of black magic allows us to match against
            # against multiple versions of the same string
            # ... etc
            match = (
                next(
                    (
                        key
                        for key in PUSHSAFER_SOUND_MAP
                        if key.startswith(sound)
                    ),
                    None,
                )
                if sound
                else None
            )

            # Now test to see if we got a match
            if not match:
                msg = f"An invalid PushSafer sound ({sound}) was specified."
                self.logger.warning(msg)
                raise TypeError(msg) from None

            # store our successfully looked up sound
            self.sound = PUSHSAFER_SOUND_MAP[match]

        if (
            self.sound is not None
            and self.sound not in PUSHSAFER_SOUND_MAP.values()
        ):
            msg = f"An invalid PushSafer sound ({sound}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg) from None

        #
        # Vibration
        #
        try:
            # Use defined integer as is if defined, no further error checking
            # is performed
            self.vibration = int(vibration)

        except TypeError:
            # NoneType means use Default; this is an okay exception
            self.vibration = None

        except ValueError:
            msg = (
                f"An invalid PushSafer vibration ({vibration}) was specified."
            )
            self.logger.warning(msg)
            raise TypeError(msg) from None

        if self.vibration and self.vibration not in PUSHSAFER_VIBRATIONS:
            msg = (
                f"An invalid PushSafer vibration ({vibration}) was specified."
            )
            self.logger.warning(msg)
            raise TypeError(msg) from None

        #
        # Private Key (associated with project)
        #
        self.privatekey = validate_regex(privatekey)
        if not self.privatekey:
            msg = (
                "An invalid PushSafer Private Key "
                f"({privatekey}) was specified."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        self.targets = parse_list(targets)
        if len(self.targets) == 0:
            self.targets = (PUSHSAFER_SEND_TO_ALL,)

        return

    def send(
        self,
        body,
        title="",
        notify_type=NotifyType.INFO,
        attach=None,
        **kwargs,
    ):
        """Perform PushSafer Notification."""

        # error tracking (used for function return)
        has_error = False

        # Initialize our list of attachments
        attachments = []

        if attach and self.attachment_support:
            # We need to upload our payload first so that we can source it
            # in remaining messages
            for no, attachment in enumerate(attach, start=1):
                # prepare payload
                if not attachment:
                    # We could not access the attachment
                    self.logger.error(
                        "Could not access PushSafer attachment"
                        f" {attachment.url(privacy=True)}."
                    )
                    return False

                if not attachment.mimetype.startswith("image/"):
                    # Attachment not supported; continue peacefully
                    self.logger.debug(
                        "Ignoring unsupported PushSafer attachment"
                        f" {attachment.url(privacy=True)}."
                    )
                    continue

                self.logger.debug(
                    "Posting PushSafer attachment"
                    f" {attachment.url(privacy=True)}"
                )

                try:
                    # Output must be in a DataURL format (that's what
                    # PushSafer calls it):
                    attachments.append((
                        (
                            attachment.name
                            if attachment.name
                            else f"file{no:03}.dat"
                        ),
                        f"data:{attachment.mimetype};base64,{attachment.base64()}",
                    ))

                except exception.AppriseException:
                    # We could not access the attachment
                    self.logger.error(
                        "Could not access PushSafer attachment"
                        f" {attachment.url(privacy=True)}."
                    )
                    return False

                self.logger.debug(
                    "Appending PushSafer attachment"
                    f" {attachment.url(privacy=True)}"
                )

        # Create a copy of the targets list
        targets = list(self.targets)
        while len(targets):
            recipient = targets.pop(0)

            # prepare payload
            payload = {
                "t": title,
                "m": body,
                # Our default icon to use
                "i": self.default_pushsafer_icon,
                # Notification Color
                "c": self.color(notify_type),
                # Target Recipient
                "d": recipient,
            }

            if self.sound is not None:
                # Only apply sound setting if it was specified
                payload["s"] = str(self.sound)

            if self.vibration is not None:
                # Only apply vibration setting
                payload["v"] = str(self.vibration)

            if not attachments:
                okay, _response = self._send(payload)
                if not okay:
                    has_error = True
                    continue

                self.logger.info(
                    f'Sent PushSafer notification to "{recipient}".'
                )

            else:
                # Create a copy of our payload object
                _payload = payload.copy()

                for idx in range(0, len(attachments), len(PICTURE_PARAMETER)):
                    # Send our attachments to our same user (already prepared
                    # as our payload object)
                    for c, attachment in enumerate(
                        attachments[idx : idx + len(PICTURE_PARAMETER)]
                    ):

                        # Get our attachment information
                        filename, dataurl = attachment
                        _payload.update({PICTURE_PARAMETER[c]: dataurl})

                        self.logger.debug(
                            f'Added attachment ({filename}) to "{recipient}".'
                        )

                    okay, _response = self._send(_payload)
                    if not okay:
                        has_error = True
                        continue

                    self.logger.info(
                        f"Sent PushSafer attachment ({filename}) to"
                        f' "{recipient}".'
                    )

                    # More then the maximum messages shouldn't cause all of
                    # the text to loop on future iterations
                    _payload = payload.copy()
                    _payload["t"] = ""
                    _payload["m"] = "..."

        return not has_error

    def _send(self, payload, **kwargs):
        """Wrapper to the requests (post) object."""

        headers = {
            "User-Agent": self.app_id,
        }

        # Prepare the notification URL to post to
        notify_url = self.notify_url.format(
            schema="https" if self.secure else "http"
        )

        # Store the payload key
        payload["k"] = self.privatekey

        self.logger.debug(
            "PushSafer POST URL:"
            f" {notify_url} (cert_verify={self.verify_certificate!r})"
        )
        self.logger.debug(f"PushSafer Payload: {payload!s}")

        # Always call throttle before any remote server i/o is made
        self.throttle()

        # Default response type
        response = None

        # Initialize our Pushsafer expected responses
        _code = None
        _str = "Unknown"

        try:
            # Open our attachment path if required:
            r = requests.post(
                notify_url,
                data=payload,
                headers=headers,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )

            try:
                response = loads(r.content)
                _code = response.get("status")
                _str = (
                    response.get("success", _str)
                    if _code == 1
                    else response.get("error", _str)
                )

            except (AttributeError, TypeError, ValueError):
                # ValueError = r.content is Unparsable
                # TypeError = r.content is None
                # AttributeError = r is None

                # Fall back to the existing unparsed value
                response = r.content

            if r.status_code not in (
                requests.codes.ok,
                requests.codes.no_content,
            ):
                # We had a problem
                status_str = NotifyPushSafer.http_response_code_lookup(
                    r.status_code
                )

                self.logger.warning(
                    "Failed to deliver payload to PushSafer:"
                    "{}{}error={}.".format(
                        status_str, ", " if status_str else "", r.status_code
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                return False, response

            elif _code != 1:
                # It's a bit backwards, but:
                #    1 is returned if we succeed
                #    0 is returned if we fail
                self.logger.warning(
                    f"Failed to deliver payload to PushSafer; error={_str}."
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                return False, response

            # otherwise we were successful
            return True, response

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred communicating with PushSafer."
            )
            self.logger.debug(f"Socket Exception: {e!s}")

            return False, response

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (
            self.secure_protocol if self.secure else self.protocol,
            self.privatekey,
        )

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Our URL parameters
        params = self.url_parameters(privacy=privacy, *args, **kwargs)

        if self.priority is not None:
            # Store our priority; but only if it was specified
            params["priority"] = next(
                (
                    key
                    for key, value in PUSHSAFER_PRIORITY_MAP.items()
                    if value == self.priority
                ),
                DEFAULT_PRIORITY,
            )  # pragma: no cover

        if self.sound is not None:
            # Store our sound; but only if it was specified
            params["sound"] = next(
                (
                    key
                    for key, value in PUSHSAFER_SOUND_MAP.items()
                    if value == self.sound
                ),
                "",
            )  # pragma: no cover

        if self.vibration is not None:
            # Store our vibration; but only if it was specified
            params["vibration"] = str(self.vibration)

        targets = "/".join([NotifyPushSafer.quote(x) for x in self.targets])
        if targets == PUSHSAFER_SEND_TO_ALL:
            # keyword is reserved for internal usage only; it's safe to remove
            # it from the recipients list
            targets = ""

        return "{schema}://{privatekey}/{targets}?{params}".format(
            schema=self.secure_protocol if self.secure else self.protocol,
            privatekey=self.pprint(self.privatekey, privacy, safe=""),
            targets=targets,
            params=NotifyPushSafer.urlencode(params),
        )

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        return len(self.targets)

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Fetch our targets
        results["targets"] = NotifyPushSafer.split_path(results["fullpath"])

        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += NotifyPushSafer.parse_list(
                results["qsd"]["to"]
            )

        # Setup the token; we store it in Private Key for global
        # plugin consistency with naming conventions
        results["privatekey"] = NotifyPushSafer.unquote(results["host"])

        if "priority" in results["qsd"] and len(results["qsd"]["priority"]):
            results["priority"] = NotifyPushSafer.unquote(
                results["qsd"]["priority"]
            )

        if "sound" in results["qsd"] and len(results["qsd"]["sound"]):
            results["sound"] = NotifyPushSafer.unquote(results["qsd"]["sound"])

        if "vibration" in results["qsd"] and len(results["qsd"]["vibration"]):
            results["vibration"] = NotifyPushSafer.unquote(
                results["qsd"]["vibration"]
            )

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from base64 import urlsafe_b64encode
import hashlib
from json import loads
from os import urandom

import requests

from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..url import PrivacyMode
from ..utils.parse import validate_regex
from .base import NotifyBase

try:
    from cryptography.hazmat.backends import default_backend
    from cryptography.hazmat.primitives import padding
    from cryptography.hazmat.primitives.ciphers import (
        Cipher,
        algorithms,
        modes,
    )

    # We're good to go!
    NOTIFY_SIMPLEPUSH_ENABLED = True

except ImportError:
    # cryptography is required in order for this package to work
    NOTIFY_SIMPLEPUSH_ENABLED = False


class NotifySimplePush(NotifyBase):
    """A wrapper for SimplePush Notifications."""

    # Set our global enabled flag
    enabled = NOTIFY_SIMPLEPUSH_ENABLED

    requirements = {
        # Define our required packaging in order to work
        "packages_required": "cryptography"
    }

    # The default descriptive name associated with the Notification
    service_name = "SimplePush"

    # The services URL
    service_url = "https://simplepush.io/"

    # The default secure protocol
    secure_protocol = "spush"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_simplepush"

    # SimplePush uses the http protocol with SimplePush requests
    notify_url = "https://api.simplepush.io/send"

    # The maximum allowable characters allowed in the body per message
    body_maxlen = 10000

    # Defines the maximum allowable characters in the title
    title_maxlen = 1024

    # Define object templates
    templates = (
        "{schema}://{apikey}",
        "{schema}://{salt}:{password}@{apikey}",
    )

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "apikey": {
                "name": _("API Key"),
                "type": "string",
                "private": True,
                "required": True,
            },
            # Used for encrypted logins
            "password": {
                "name": _("Password"),
                "type": "string",
                "private": True,
            },
            "salt": {
                "name": _("Salt"),
                "type": "string",
                "private": True,
                "map_to": "user",
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "event": {
                "name": _("Event"),
                "type": "string",
            },
        },
    )

    def __init__(self, apikey, event=None, **kwargs):
        """Initialize SimplePush Object."""
        super().__init__(**kwargs)

        # API Key (associated with project)
        self.apikey = validate_regex(apikey)
        if not self.apikey:
            msg = f"An invalid SimplePush API Key ({apikey}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        if event:
            # Event Name (associated with project)
            self.event = validate_regex(event)
            if not self.event:
                msg = (
                    "An invalid SimplePush Event Name "
                    f"({event}) was specified."
                )
                self.logger.warning(msg)
                raise TypeError(msg)

        else:
            # Default Event Name
            self.event = None

        # Used/cached in _encrypt() function
        self._iv = None
        self._iv_hex = None
        self._key = None

    def _encrypt(self, content):
        """Encrypts message for use with SimplePush."""

        if self._iv is None:
            # initialization vector and cache it
            self._iv = urandom(algorithms.AES.block_size // 8)

            # convert vector into hex string (used in payload)
            self._iv_hex = "".join([
                f"{ord(self._iv[idx:idx + 1]):02x}"
                for idx in range(len(self._iv))
            ]).upper()

            # encrypted key and cache it
            self._key = bytes(
                bytearray.fromhex(
                    hashlib.sha1(
                        f"{self.password}{self.user}".encode()
                    ).hexdigest()[0:32]
                )
            )

        padder = padding.PKCS7(algorithms.AES.block_size).padder()
        content = padder.update(content.encode()) + padder.finalize()
        #
        # Encryption Notice
        #

        # CBC mode doesn't provide integrity guarantees. Unless the message
        # authentication for IV and the ciphertext are applied, it will be
        # vulnerable to a padding oracle attack

        # It is important to identify that both the Apprise package and team
        # recognizes this AES-CBC-128 weakness but requires that it exists due
        # to it being the SimplePush Requirement as documented on their
        # website here https://simplepush.io/features.

        # In the event the website link above does not exist/work, a screen
        # capture of the reference to the requirement for this encryption
        # can also be found on the Apprise SimplePush Wiki:
        #   https://github.com/caronc/apprise/wiki/Notify_simplepush\
        #       #lock-aes-cbc-128-encryption-weakness
        #
        encryptor = Cipher(
            algorithms.AES(self._key), modes.CBC(self._iv), default_backend()
        ).encryptor()

        return urlsafe_b64encode(
            encryptor.update(content) + encryptor.finalize()
        )

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform SimplePush Notification."""

        headers = {
            "User-Agent": self.app_id,
            "Content-type": "application/x-www-form-urlencoded",
        }

        # Prepare our payload
        payload = {
            "key": self.apikey,
        }

        if self.password and self.user:
            body = self._encrypt(body)
            title = self._encrypt(title)
            payload.update({
                "encrypted": "true",
                "iv": self._iv_hex,
            })

        # prepare SimplePush Object
        payload.update({
            "msg": body,
            "title": title,
        })

        if self.event:
            # Store Event
            payload["event"] = self.event

        self.logger.debug(
            "SimplePush POST URL:"
            f" {self.notify_url} (cert_verify={self.verify_certificate!r})"
        )
        self.logger.debug(f"SimplePush Payload: {payload!s}")

        # We need to rely on the status string returned in the SimplePush
        # response
        status_str = None
        status = None

        # Always call throttle before any remote server i/o is made
        self.throttle()

        try:
            r = requests.post(
                self.notify_url,
                data=payload,
                headers=headers,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )

            # Get our SimplePush response (if it's possible)
            try:
                json_response = loads(r.content)
                status_str = json_response.get("message")
                status = json_response.get("status")

            except (TypeError, ValueError, AttributeError):
                # TypeError = r.content is not a String
                # ValueError = r.content is Unparsable
                # AttributeError = r.content is None
                pass

            if r.status_code != requests.codes.ok or status != "OK":
                # We had a problem
                status_str = (
                    status_str
                    if status_str
                    else NotifyBase.http_response_code_lookup(r.status_code)
                )

                self.logger.warning(
                    "Failed to send SimplePush notification:"
                    "{}{}error={}.".format(
                        status_str, ", " if status_str else "", r.status_code
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                # Return; we're done
                return False

            else:
                self.logger.info("Sent SimplePush notification.")

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred sending SimplePush notification."
            )
            self.logger.debug(f"Socket Exception: {e!s}")

            # Return; we're done
            return False

        return True

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (self.secure_protocol, self.user, self.password, self.apikey)

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Our URL parameters
        params = self.url_parameters(privacy=privacy, *args, **kwargs)

        if self.event:
            params["event"] = self.event

        # Determine Authentication
        auth = ""
        if self.user and self.password:
            auth = "{salt}:{password}@".format(
                salt=self.pprint(
                    self.user, privacy, mode=PrivacyMode.Secret, safe=""
                ),
                password=self.pprint(
                    self.password, privacy, mode=PrivacyMode.Secret, safe=""
                ),
            )

        return "{schema}://{auth}{apikey}/?{params}".format(
            schema=self.secure_protocol,
            auth=auth,
            apikey=self.pprint(self.apikey, privacy, safe=""),
            params=NotifySimplePush.urlencode(params),
        )

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Set the API Key
        results["apikey"] = NotifySimplePush.unquote(results["host"])

        # Event
        if "event" in results["qsd"] and len(results["qsd"]["event"]):
            # Extract the account sid from an argument
            results["event"] = NotifySimplePush.unquote(
                results["qsd"]["event"]
            )

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import contextlib
from itertools import chain
from json import dumps
import os
import time

import requests

from ...common import NotifyImageSize, NotifyType, PersistentStoreMode
from ...locale import gettext_lazy as _
from ...utils import pem as _pem
from ...utils.base64 import base64_urlencode
from ...utils.parse import is_email, parse_bool, parse_list
from ..base import NotifyBase
from . import subscription


class VapidPushMode:
    """Supported Vapid Push Services."""

    CHROME = "chrome"
    FIREFOX = "firefox"
    EDGE = "edge"
    OPERA = "opera"
    APPLE = "apple"
    SAMSUNG = "samsung"
    BRAVE = "brave"
    GENERIC = "generic"


VAPID_API_LOOKUP = {
    VapidPushMode.CHROME: "https://fcm.googleapis.com/fcm/send",
    VapidPushMode.FIREFOX: (
        "https://updates.push.services.mozilla.com/wpush/v1"
    ),
    VapidPushMode.EDGE: (
        "https://fcm.googleapis.com/fcm/send"
    ),  # Edge uses FCM too
    VapidPushMode.OPERA: (
        "https://fcm.googleapis.com/fcm/send"
    ),  # Opera is Chromium-based
    VapidPushMode.APPLE: (
        "https://web.push.apple.com"
    ),  # Apple Web Push base endpoint
    VapidPushMode.BRAVE: "https://fcm.googleapis.com/fcm/send",
    VapidPushMode.SAMSUNG: "https://fcm.googleapis.com/fcm/send",
    VapidPushMode.GENERIC: "https://fcm.googleapis.com/fcm/send",
}

VAPID_PUSH_MODES = (
    VapidPushMode.CHROME,
    VapidPushMode.FIREFOX,
    VapidPushMode.EDGE,
    VapidPushMode.OPERA,
    VapidPushMode.APPLE,
)


class NotifyVapid(NotifyBase):
    """A wrapper for WebPush/Vapid notifications."""

    # Set our global enabled flag
    enabled = subscription.CRYPTOGRAPHY_SUPPORT and _pem.PEM_SUPPORT

    requirements = {
        # Define our required packaging in order to work
        "packages_required": "cryptography"
    }

    # The default descriptive name associated with the Notification
    service_name = "Vapid Web Push Notifications"

    # The services URL
    service_url = (
        "https://datatracker.ietf.org/doc/html/draft-thomson-webpush-vapid"
    )

    # The default protocol
    secure_protocol = "vapid"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_vapid"

    # There is no reason we should exceed 5KB when reading in a PEM file.
    # If it is more than this, then it is not accepted.
    max_vapid_keyfile_size = 5000

    # There is no reason we should exceed 5MB when reading in a JSON file.
    # If it is more than this, then it is not accepted.
    max_vapid_subfile_size = 5242880

    # The maximum length of the messge can be 4096
    # just choosing a safe number below this to allow for padding and
    # encryption
    body_maxlen = 4000

    # A title can not be used for SMS Messages.  Setting this to zero will
    # cause any title (if defined) to get placed into the message body.
    title_maxlen = 0

    # Our default is to no not use persistent storage beyond in-memory
    # reference; this allows us to auto-generate our config if needed
    storage_mode = PersistentStoreMode.AUTO

    # 43200 = 12 hours
    vapid_jwt_expiration_sec = 43200

    # Subscription file
    vapid_subscription_file = "subscriptions.json"

    # Allows the user to specify the NotifyImageSize object
    image_size = NotifyImageSize.XY_72

    # Define object templates
    templates = (
        "{schema}://{subscriber}",
        "{schema}://{subscriber}/{targets}",
    )

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "subscriber": {
                "name": _("API Key"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
            },
        },
    )

    # Define our template args
    template_args = dict(
        NotifyBase.template_tokens,
        **{
            "mode": {
                "name": _("Mode"),
                "type": "choice:string",
                "values": VAPID_PUSH_MODES,
                "default": VAPID_PUSH_MODES[0],
                "map_to": "mode",
            },
            # Default Time To Live (defined in seconds)
            # 0 (Zero) - message will be delivered only if the device is
            # reacheable
            "ttl": {
                "name": _("ttl"),
                "type": "int",
                "default": 0,
                "min": 0,
                "max": 60,
            },
            "to": {
                "alias_of": "targets",
            },
            "from": {
                "alias_of": "subscriber",
            },
            "keyfile": {
                # A Private Keyfile is required to sign header
                "name": _("PEM Private KeyFile"),
                "type": "string",
                "private": True,
            },
            "subfile": {
                # A Subscripion File is required to sign header
                "name": _("Subscripion File"),
                "type": "string",
                "private": True,
            },
            "image": {
                "name": _("Include Image"),
                "type": "bool",
                "default": True,
                "map_to": "include_image",
            },
        },
    )

    def __init__(
        self,
        subscriber,
        mode=None,
        targets=None,
        keyfile=None,
        subfile=None,
        include_image=None,
        ttl=None,
        **kwargs,
    ):
        """Initialize Vapid Messaging."""
        super().__init__(**kwargs)

        # Path to our Private Key file
        self.keyfile = None

        # Path to our subscription.json file
        self.subfile = None

        #
        # Our Targets
        #
        self.targets = []
        self._invalid_targets = []

        # default subscriptions
        self.subscriptions = {}
        self.subscriptions_loaded = False
        self.private_key_loaded = False

        # Set our Time to Live Flag
        self.ttl = self.template_args["ttl"]["default"]
        if ttl is not None:
            with contextlib.suppress(ValueError, TypeError):
                # Store our TTL (Time To live) if it is a valid integer
                self.ttl = int(ttl)

            if (
                self.ttl < self.template_args["ttl"]["min"]
                or self.ttl > self.template_args["ttl"]["max"]
            ):
                msg = f"The Vapid TTL specified ({self.ttl}) is out of range."
                self.logger.warning(msg)
                raise TypeError(msg)

        # Place a thumbnail image inline with the message body
        self.include_image = (
            self.template_args["image"]["default"]
            if include_image is None
            else include_image
        )

        result = is_email(subscriber)
        if not result:
            msg = f"An invalid Vapid Subscriber({subscriber}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)
        self.subscriber = result["full_email"]

        # Store our Mode/service
        try:
            self.mode = (
                NotifyVapid.template_args["mode"]["default"]
                if mode is None
                else mode.lower()
            )

            if self.mode not in VAPID_PUSH_MODES:
                # allow the outer except to handle this common response
                raise IndexError()

        except (AttributeError, IndexError, TypeError):
            # Invalid region specified
            msg = f"The Vapid mode specified ({mode}) is invalid."
            self.logger.warning(msg)
            raise TypeError(msg) from None

        # Our Private keyfile
        self.keyfile = keyfile

        # Our Subscription file
        self.subfile = subfile

        # Prepare our PEM Object
        self.pem = _pem.ApprisePEMController(self.store.path, asset=self.asset)

        # Create our subscription object
        self.subscriptions = subscription.WebPushSubscriptionManager(
            asset=self.asset
        )

        if (
            self.subfile is None
            and self.store.mode != PersistentStoreMode.MEMORY
            and self.asset.pem_autogen
        ):

            self.subfile = os.path.join(
                self.store.path, self.vapid_subscription_file
            )
            if not os.path.exists(self.subfile) and self.subscriptions.write(
                self.subfile
            ):
                self.logger.info(
                    "Vapid auto-generated %s/%s",
                    os.path.basename(self.store.path),
                    self.vapid_subscription_file,
                )

        # Acquire our targets for parsing
        self.targets = parse_list(targets)
        if not self.targets:
            # Add ourselves
            self.targets.append(self.subscriber)

        return

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform Vapid Notification."""
        if not self.private_key_loaded and (
            (
                self.keyfile
                and not self.pem.private_key(autogen=False, autodetect=False)
                and not self.pem.load_private_key(self.keyfile)
            )
            or (not self.keyfile and not self.pem)
        ):
            self.logger.warning(
                "Provided Vapid/WebPush (PEM) Private Key file could "
                "not be loaded."
            )
            self.private_key_loaded = True
            return False
        else:
            self.private_key_loaded = True

        if not self.targets:
            # There is no one to notify; we're done
            self.logger.warning("There are no Vapid targets to notify")
            return False

        if not self.subscriptions_loaded and self.subfile:
            # Toggle our loaded flag to prevent trying again later
            self.subscriptions_loaded = True
            if not self.subscriptions.load(
                self.subfile, byte_limit=self.max_vapid_subfile_size
            ):
                self.logger.warning(
                    "Provided Vapid/WebPush subscriptions file could not be "
                    "loaded."
                )
                return False

        if not self.subscriptions:
            self.logger.warning("Vapid could not load subscriptions")
            return False

        if not self.pem.private_key(autogen=False, autodetect=False):
            self.logger.warning(
                "No Vapid/WebPush (PEM) Private Key file could be loaded."
            )
            return False

        # Prepare our notify URL (based on our mode)
        notify_url = VAPID_API_LOOKUP[self.mode]
        headers = {
            "User-Agent": self.app_id,
            "TTL": str(self.ttl),
            "Content-Encoding": "aes128gcm",
            "Content-Type": "application/octet-stream",
            "Authorization": f"vapid t={self.jwt_token}, k={self.public_key}",
        }

        has_error = False

        # Create a copy of the targets list
        targets = list(self.targets)
        while len(targets):
            target = targets.pop(0)
            if target not in self.subscriptions:
                self.logger.warning(
                    "Dropped Vapid user "
                    f"({target}) specified - not found in subscriptions.json.",
                )
                # Save ourselves from doing this again
                self._invalid_targets.append(target)
                self.targets.remove(target)
                has_error = True
                continue

            # Encrypt our payload
            encrypted_payload = self.pem.encrypt_webpush(
                body,
                public_key=self.subscriptions[target].public_key,
                auth_secret=self.subscriptions[target].auth_secret,
            )

            self.logger.debug(
                "Vapid %s POST URL: %s (cert_verify=%r)",
                self.mode,
                notify_url,
                self.verify_certificate,
            )
            self.logger.debug(
                "Vapid %s Encrypted Payload: %d byte(s)", self.mode, len(body)
            )

            # Always call throttle before any remote server i/o is made
            self.throttle()
            try:
                r = requests.post(
                    notify_url,
                    data=encrypted_payload,
                    headers=headers,
                    verify=self.verify_certificate,
                    timeout=self.request_timeout,
                )
                if r.status_code not in (
                    requests.codes.ok,
                    requests.codes.no_content,
                ):
                    # We had a problem
                    status_str = NotifyBase.http_response_code_lookup(
                        r.status_code
                    )

                    self.logger.warning(
                        "Failed to send {} Vapid notification: "
                        "{}{}error={}.".format(
                            self.mode,
                            status_str,
                            ", " if status_str else "",
                            r.status_code,
                        )
                    )

                    self.logger.debug("Response Details:\r\n%s", r.content)

                    has_error = True

                else:
                    self.logger.info("Sent %s Vapid notification.", self.mode)

            except requests.RequestException as e:
                self.logger.warning(
                    "A Connection error occurred sending Vapid notification."
                )
                self.logger.debug("Socket Exception: %s", str(e))

                has_error = True

        return not has_error

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (self.secure_protocol, self.mode, self.subscriber)

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "mode": self.mode,
            "ttl": str(self.ttl),
        }

        if self.keyfile:
            # Include our keyfile if specified
            params["keyfile"] = self.keyfile

        if self.subfile:
            # Include our subfile if specified
            params["subfile"] = self.subfile

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        targets = (
            self.targets
            if not (
                self.targets == 1
                and self.targets[0].lower() == self.subscriber.lower()
            )
            else []
        )
        return "{schema}://{subscriber}/{targets}?{params}".format(
            schema=self.secure_protocol,
            subscriber=NotifyVapid.quote(self.subscriber, safe="@"),
            targets="/".join(
                chain(
                    [str(t) for t in targets],
                    [
                        NotifyVapid.quote(x, safe="@")
                        for x in self._invalid_targets
                    ],
                )
            ),
            params=NotifyVapid.urlencode(params),
        )

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        targets = len(self.targets)
        return targets if targets else 1

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Prepare our targets
        results["targets"] = []
        if "from" in results["qsd"] and len(results["qsd"]["from"]):
            results["subscriber"] = NotifyVapid.unquote(results["qsd"]["from"])

            if results["user"] and results["host"]:
                # whatever is left on the URL goes
                results["targets"].append(
                    "{}@{}".format(
                        NotifyVapid.unquote(results["user"]),
                        NotifyVapid.unquote(results["host"]),
                    )
                )

            elif results["host"]:
                results["targets"].append(NotifyVapid.unquote(results["host"]))

        else:
            # Acquire our subscriber information
            results["subscriber"] = "{}@{}".format(
                NotifyVapid.unquote(results["user"]),
                NotifyVapid.unquote(results["host"]),
            )

        results["targets"].extend(NotifyVapid.split_path(results["fullpath"]))

        # Get our mode
        results["mode"] = results["qsd"].get("mode")

        # Get Image Flag
        results["include_image"] = parse_bool(
            results["qsd"].get(
                "image", NotifyVapid.template_args["image"]["default"]
            )
        )

        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += NotifyVapid.parse_list(results["qsd"]["to"])

        # Our Private Keyfile (PEM)
        if "keyfile" in results["qsd"] and results["qsd"]["keyfile"]:
            results["keyfile"] = NotifyVapid.unquote(results["qsd"]["keyfile"])

        # Our Subscription File (JSON)
        if "subfile" in results["qsd"] and results["qsd"]["subfile"]:
            results["subfile"] = NotifyVapid.unquote(results["qsd"]["subfile"])

        # Support the 'ttl' variable
        if "ttl" in results["qsd"] and len(results["qsd"]["ttl"]):
            results["ttl"] = NotifyVapid.unquote(results["qsd"]["ttl"])

        return results

    @property
    def jwt_token(self):
        """Returns our VAPID Token based on class details."""
        # JWT header
        header = {"alg": "ES256", "typ": "JWT"}

        # JWT payload
        payload = {
            "aud": VAPID_API_LOOKUP[self.mode],
            "exp": int(time.time()) + self.vapid_jwt_expiration_sec,
            "sub": f"mailto:{self.subscriber}",
        }

        # Base64 URL encode header and payload
        header_b64 = base64_urlencode(
            dumps(header, separators=(",", ":")).encode("utf-8")
        )
        payload_b64 = base64_urlencode(
            dumps(payload, separators=(",", ":")).encode("utf-8")
        )
        signing_input = f"{header_b64}.{payload_b64}".encode()
        signature_b64 = base64_urlencode(self.pem.sign(signing_input))

        # Return final token
        return f"{header_b64}.{payload_b64}.{signature_b64}"

    @property
    def public_key(self):
        """Returns our public key representation."""
        return self.pem.x962_str



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Simple API Reference:
#  - https://sendpulse.com/integrations/api/smtp

import base64
from email.utils import formataddr
from json import dumps, loads
import re

import requests

from .. import exception
from ..common import NotifyFormat, NotifyType, PersistentStoreMode
from ..conversion import convert_between
from ..locale import gettext_lazy as _
from ..utils.parse import is_email, parse_emails, validate_regex
from .base import NotifyBase


class NotifySendPulse(NotifyBase):
    """
    A wrapper for Notify SendPulse Notifications
    """

    # The default descriptive name associated with the Notification
    service_name = "SendPulse"

    # The services URL
    service_url = "https://sendpulse.com"

    # The default secure protocol
    secure_protocol = "sendpulse"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_sendpulse"

    # Default to markdown
    notify_format = NotifyFormat.HTML

    # The default Email API URL to use
    notify_email_url = "https://api.sendpulse.com/smtp/emails"

    # Our OAuth Query
    notify_oauth_url = "https://api.sendpulse.com/oauth/access_token"

    # Support attachments
    attachment_support = True

    # Allow 300 requests per minute.
    # 60/300 = 0.2
    request_rate_per_sec = 0.2

    # Our default is to no not use persistent storage beyond in-memory
    # reference
    storage_mode = PersistentStoreMode.AUTO

    # Token expiry if not detected in seconds (below is 1 hr)
    token_expiry = 3600

    # The number of seconds to grace for early token renewal
    # Below states that 10 seconds bfore our token expiry, we'll
    # attempt to renew it
    token_expiry_edge = 10

    # Support attachments
    attachment_support = True

    # The default subject to use if one isn't specified.
    default_empty_subject = "<no subject>"

    # Define object templates
    templates = (
        "{schema}://{user}@{host}/{client_secret}/",
        "{schema}://{user}@{host}/{client_id}/{client_secret}/{targets}",
    )

    # Define our template arguments
    template_tokens = dict(NotifyBase.template_tokens, **{
        "user": {
            "name": _("User Name"),
            "type": "string",
        },
        "host": {
            "name": _("Domain"),
            "type": "string",
            "required": True,
        },
        "client_id": {
            "name": _("Client ID"),
            "type": "string",
            "required": True,
            "private": True,
            "regex": (r"^[A-Z0-9._-]+$", "i"),
        },
        "client_secret": {
            "name": _("Client Secret"),
            "type": "string",
            "required": True,
            "private": True,
            "regex": (r"^[A-Z0-9._-]+$", "i"),
        },
        "target_email": {
            "name": _("Target Email"),
            "type": "string",
            "map_to": "targets",
        },
        "targets": {
            "name": _("Targets"),
            "type": "list:string",
        },
    })

    # Define our template arguments
    template_args = dict(NotifyBase.template_args, **{
        "to": {
            "alias_of": "targets",
        },
        "from": {
            "name": _("From Email"),
            "type": "string",
            "map_to": "from_addr",
        },
        "cc": {
            "name": _("Carbon Copy"),
            "type": "list:string",
        },
        "bcc": {
            "name": _("Blind Carbon Copy"),
            "type": "list:string",
        },
        "template": {
            # The template ID is an integer
            "name": _("Template ID"),
            "type": "int",
        },
        "id": {
            "alias_of": "client_id",
        },
        "secret": {
            "alias_of": "client_secret",
        }
    })

    # Support Template Dynamic Variables (Substitutions)
    template_kwargs = {
        "template_data": {
            "name": _("Template Data"),
            "prefix": "+",
        },
    }

    def __init__(self, client_id, client_secret, from_addr=None, targets=None,
                 cc=None, bcc=None, template=None, template_data=None,
                 **kwargs):
        """
        Initialize Notify SendPulse Object
        """
        super().__init__(**kwargs)

        # For tracking our email -> name lookups
        self.names = {}

        # Temporary from_addr to work with for parsing
        _from_addr = [self.app_id, ""]

        if self.user:
            if self.host:
                # Prepare the bases of our email
                _from_addr = [_from_addr[0], "{}@{}".format(
                    re.split(r"[\s@]+", self.user)[0],
                    self.host,
                )]

            else:
                result = is_email(self.user)
                if result:
                    # Prepare the bases of our email and include domain
                    self.host = result["domain"]
                    _from_addr = [
                        result["name"] if result["name"]
                        else _from_addr[0], self.user]

        if isinstance(from_addr, str):
            result = is_email(from_addr)
            if result:
                _from_addr = (
                    result["name"] if result["name"] else _from_addr[0],
                    result["full_email"])
            else:
                # Only update the string but use the already detected info
                _from_addr[0] = from_addr

        result = is_email(_from_addr[1])
        if not result:
            # Parse Source domain based on from_addr
            msg = "Invalid ~From~ email specified: {}".format(
                "{} <{}>".format(_from_addr[0], _from_addr[1])
                if _from_addr[0] else "{}".format(_from_addr[1]))
            self.logger.warning(msg)
            raise TypeError(msg)

        # Store our lookup
        self.from_addr = _from_addr[1]
        self.names[_from_addr[1]] = _from_addr[0]

        # Client ID
        self.client_id = validate_regex(
            client_id, *self.template_tokens["client_id"]["regex"])
        if not self.client_id:
            msg = "An invalid SendPulse Client ID " \
                  "({}) was specified.".format(client_id)
            self.logger.warning(msg)
            raise TypeError(msg)

        # Client Secret
        self.client_secret = validate_regex(
            client_secret, *self.template_tokens["client_secret"]["regex"])
        if not self.client_secret:
            msg = "An invalid SendPulse Client Secret " \
                  "({}) was specified.".format(client_secret)
            self.logger.warning(msg)
            raise TypeError(msg)

        # Acquire Targets (To Emails)
        self.targets = []

        # Acquire Carbon Copies
        self.cc = set()

        # Acquire Blind Carbon Copies
        self.bcc = set()

        # No template
        self.template = None
        if template:
            try:
                # Store our template
                self.template = int(template)

            except (TypeError, ValueError):
                # Not a valid integer; ignore entry
                err = "The SendPulse Template ID specified ({}) is invalid."\
                    .format(template)
                self.logger.warning(err)
                raise TypeError(err) from None

        # Now our dynamic template data (if defined)
        self.template_data = template_data \
            if isinstance(template_data, dict) else {}

        if targets:
            # Validate recipients (to:) and drop bad ones:
            for recipient in parse_emails(targets):
                result = is_email(recipient)
                if result:
                    self.targets.append(result["full_email"])
                    if result["name"]:
                        self.names[result["full_email"]] = result["name"]
                    continue

                self.logger.warning(
                    "Dropped invalid To email "
                    "({}) specified.".format(recipient),
                )

        else:
            # If our target email list is empty we want to add ourselves to it
            self.targets.append(self.from_addr)

        # Validate recipients (cc:) and drop bad ones:
        for recipient in parse_emails(cc):

            result = is_email(recipient)
            if result:
                self.cc.add(result["full_email"])
                if result["name"]:
                    self.names[result["full_email"]] = result["name"]
                continue

            self.logger.warning(
                "Dropped invalid Carbon Copy email "
                "({}) specified.".format(recipient),
            )

        # Validate recipients (bcc:) and drop bad ones:
        for recipient in parse_emails(bcc):

            result = is_email(recipient)
            if result:
                self.bcc.add(result["full_email"])
                if result["name"]:
                    self.names[result["full_email"]] = result["name"]
                continue

            self.logger.warning(
                "Dropped invalid Blind Carbon Copy email "
                "({}) specified.".format(recipient),
            )

        if len(self.targets) == 0:
            # Notify ourselves
            self.targets.append(self.from_addr)

        return

    @property
    def url_identifier(self):
        """
        Returns all of the identifiers that make this URL unique from
        another simliar one. Targets or end points should never be identified
        here.
        """
        return (self.secure_protocol, self.client_id, self.client_secret)

    def url(self, privacy=False, *args, **kwargs):
        """
        Returns the URL built dynamically based on specified arguments.
        """

        # Our URL parameters
        params = self.url_parameters(privacy=privacy, *args, **kwargs)

        if len(self.cc) > 0:
            # Handle our Carbon Copy Addresses
            params["cc"] = ",".join([
                formataddr(
                    (self.names.get(e, False), e),
                    # Swap comma for it's escaped url code (if detected) since
                    # we're using that as a delimiter
                    charset="utf-8").replace(",", "%2C")
                for e in self.cc])

        if len(self.bcc) > 0:
            # Handle our Blind Carbon Copy Addresses
            params["bcc"] = ",".join([
                formataddr(
                    (self.names.get(e, False), e),
                    # Swap comma for it's escaped url code (if detected) since
                    # we're using that as a delimiter
                    charset="utf-8").replace(",", "%2C")
                for e in self.bcc])

        if self.template:
            # Handle our Template ID if if was specified
            params["template"] = self.template

        # handle from=
        if self.names[self.from_addr] != self.app_id:
            params["from"] = self.names[self.from_addr]

        # Append our template_data into our parameter list
        params.update(
            {"+{}".format(k): v for k, v in self.template_data.items()})

        # a simple boolean check as to whether we display our target emails
        # or not
        has_targets = \
            not (len(self.targets) == 1 and self.targets[0] == self.from_addr)

        return "{schema}://{source}/{cid}/{secret}/{targets}?{params}".format(
            schema=self.secure_protocol,
            source=self.from_addr,
            cid=self.pprint(self.client_id, privacy, safe=""),
            secret=self.pprint(self.client_secret, privacy, safe=""),
            targets="" if not has_targets else "/".join(
                [NotifySendPulse.quote(x, safe="") for x in self.targets]),
            params=NotifySendPulse.urlencode(params),
        )

    def __len__(self):
        """
        Returns the number of targets associated with this notification
        """
        return len(self.targets)

    def login(self):
        """
        Authenticates with the server to get a access_token
        """
        self.store.clear("access_token")
        payload = {
            "grant_type": "client_credentials",
            "client_id": self.client_id,
            "client_secret": self.client_secret,
        }

        success, response = self._fetch(self.notify_oauth_url, payload)
        if not success:
            return False

        access_token = response.get("access_token")

        # If we get here, we're authenticated
        try:
            expires = \
                int(response.get("expires_in")) - self.token_expiry_edge
            if expires <= self.token_expiry_edge:
                self.logger.error(
                    "SendPulse token expiry limit returned was invalid")
                return False

            elif expires > self.token_expiry:
                self.logger.warning(
                    "SendPulse token expiry limit fixed to: {}s"
                    .format(self.token_expiry))
                expires = self.token_expiry - self.token_expiry_edge

        except (AttributeError, TypeError, ValueError):
            # expires_in was not an integer
            self.logger.warning(
                "SendPulse token expiry limit presumed to be: {}s".format(
                    self.token_expiry))
            expires = self.token_expiry - self.token_expiry_edge

        self.store.set("access_token", access_token, expires=expires)

        return access_token

    def send(self, body, title="", notify_type=NotifyType.INFO, attach=None,
             **kwargs):
        """
        Perform SendPulse Notification
        """

        access_token = self.store.get("access_token") or self.login()
        if not access_token:
            return False

        # error tracking (used for function return)
        has_error = False

        # A Simple Email Payload Template
        _payload = {
            "email": {
                "from": {
                    "name": self.names[self.from_addr],
                    "email": self.from_addr,
                },
                # To is populated further on
                "to": [],
                # A subject is a requirement, so if none is specified we must
                # set a default with at least 1 character or SendPulse will
                # deny our request
                "subject": title if title else self.default_empty_subject,
            }
        }

        # Prepare Email Message
        if self.notify_format == NotifyFormat.HTML:
            # HTML
            _payload["email"].update({
                "text": convert_between(
                    NotifyFormat.HTML, NotifyFormat.TEXT, body),
                "html": base64.b64encode(body.encode("utf-8")).decode("ascii"),
            })

        else:  # Text
            _payload["email"]["text"] = body

        if attach and self.attachment_support:
            attachments = {}

            # Send our attachments
            for no, attachment in enumerate(attach, start=1):
                # Perform some simple error checking
                if not attachment:
                    # We could not access the attachment
                    self.logger.error(
                        "Could not access SendPulse attachment {}.".format(
                            attachment.url(privacy=True)))
                    return False

                try:
                    attachments[
                        attachment.name if attachment.name
                        else f"file{no:03}.dat"] = attachment.base64()

                except exception.AppriseException:
                    # We could not access the attachment
                    self.logger.error(
                        "Could not access SendPulse attachment {}.".format(
                            attachment.url(privacy=True)))
                    return False

                self.logger.debug(
                    "Appending SendPulse attachment {}".format(
                        attachment.url(privacy=True)))

            # Append our attachments to the payload
            _payload["email"].update({
                "attachments_binary": attachments,
            })

        if self.template:
            _payload["email"].update({
                "template": {
                    "id": self.template,
                    "variables": self.template_data,
                }})

        targets = list(self.targets)
        while len(targets) > 0:
            target = targets.pop(0)

            # Create a copy of our template
            payload = _payload.copy()

            # the cc, bcc, to field must be unique or SendMail will fail, the
            # below code prepares this by ensuring the target isn't in the cc
            # list or bcc list. It also makes sure the cc list does not contain
            # any of the bcc entries
            cc = (self.cc - self.bcc - {target})
            bcc = (self.bcc - {target})

            #
            # prepare our 'to'
            #
            to = {
                "email": target
            }
            if target in self.names:
                to["name"] = self.names[target]

            # Set our target
            payload["email"]["to"] = [to]

            if len(cc):
                payload["email"]["cc"] = []
                for email in cc:
                    item = {
                        "email": email,
                    }
                if email in self.names:
                    item["name"] = self.names[email]

                payload["email"]["cc"].append(item)

            if len(bcc):
                payload["email"]["bcc"] = []
                for email in bcc:
                    item = {
                        "email": email,
                    }
                if email in self.names:
                    item["name"] = self.names[email]

                payload["email"]["bcc"].append(item)

            # Perform our post
            success, _response = self._fetch(
                self.notify_email_url, payload, target, retry=1)
            if not success:
                has_error = True
                continue

        return not has_error

    def _fetch(self, url, payload, target=None, retry=0):
        """
        Wrapper to request.post() to manage it's response better and make
        the send() function cleaner and easier to maintain.

        This function returns True if the _post was successful and False
        if it wasn't.
        """
        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/json",
        }

        access_token = self.store.get("access_token")
        if access_token:
            headers.update({"Authorization": f"Bearer {access_token}"})

        self.logger.debug("SendPulse POST URL: {} (cert_verify={!r})".format(
            url, self.verify_certificate,
        ))
        self.logger.debug("SendPulse Payload: {}".format(str(payload)))

        # Prepare our default response
        response = {}

        # Always call throttle before any remote server i/o is made
        self.throttle()
        try:
            r = requests.post(
                url,
                data=dumps(payload),
                headers=headers,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )

            try:
                response = loads(r.content)

            except (AttributeError, TypeError, ValueError):
                # This gets thrown if we can't parse our JSON Response
                #  - ValueError = r.content is Unparsable
                #  - TypeError = r.content is None
                #  - AttributeError = r is None
                self.logger.warning("Invalid response from SendPulse server.")
                self.logger.debug(
                    "Response Details:\r\n{}".format(r.content))
                return (False, {})

            # Reference status code
            status_code = r.status_code

            # Key likely expired, we'll reset it and try one more time
            if status_code == requests.codes.unauthorized \
                    and retry and self.login():
                return self._fetch(url, payload, target, retry=retry - 1)

            if status_code not in (
                    requests.codes.ok, requests.codes.accepted):
                # We had a problem
                status_str = \
                    NotifySendPulse.http_response_code_lookup(
                        status_code)

                if target:
                    self.logger.warning(
                        "Failed to send SendPulse notification to {}: "
                        "{}{}error={}.".format(
                            target,
                            status_str,
                            ", " if status_str else "",
                            status_code))
                else:
                    self.logger.warning(
                        "SendPulse Authentication Request failed: "
                        "{}{}error={}.".format(
                            status_str,
                            ", " if status_str else "",
                            status_code))

                self.logger.debug(
                    "Response Details:\r\n{}".format(r.content))

            else:
                if target:
                    self.logger.info(
                        "Sent SendPulse notification to {}.".format(target))
                else:
                    self.logger.debug("SendPulse authentication successful")

                return (True, response)

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred sending SendPulse "
                "notification to {}.".format(target))
            self.logger.debug("Socket Exception: {}".format(str(e)))

        return (False, response)

    @staticmethod
    def parse_url(url):
        """
        Parses the URL and returns enough arguments that can allow
        us to re-instantiate this object.

        """

        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Define our minimum requirements; defining them now saves us from
        # having to if/else all kinds of branches below...
        results["from_addr"] = None
        results["client_id"] = None
        results["client_secret"] = None

        # Prepare our targets
        results["targets"] = []

        # Our URL looks like this:
        #    {schema}://{from_addr}:{client_id}/{client_secret}/{targets}
        #
        # which actually equates to:
        #    {schema}://{user}@{host}/{client_id}/{client_secret}
        #                                  /{email1}/{email2}/etc..
        #                 ^       ^
        #                 |       |
        #                -from addr-
        if "from" in results["qsd"]:
            results["from_addr"] = \
                NotifySendPulse.unquote(results["qsd"]["from"].rstrip())

            if is_email(results["from_addr"]):
                # Our hostname is free'd up to be interpreted as part of the
                # targets
                results["targets"].append(
                    NotifySendPulse.unquote(results["host"]))
                results["host"] = ""

        if "user" in results["qsd"] and \
                is_email(NotifySendPulse.unquote(results["user"])):
            # Our hostname is free'd up to be interpreted as part of the
            # targets
            results["targets"].append(NotifySendPulse.unquote(results["host"]))
            results["host"] = ""

        # Get our potential email targets
        # First 2 elements are the client_id and client_secret
        results["targets"] += NotifySendPulse.split_path(results["fullpath"])
        # check for our client id
        if "id" in results["qsd"] and len(results["qsd"]["id"]):
            # Store our Client ID
            results["client_id"] = \
                NotifySendPulse.unquote(results["qsd"]["id"])

        elif results["targets"]:
            # Store our Client ID
            results["client_id"] = results["targets"].pop(0)

        if "secret" in results["qsd"] and len(results["qsd"]["secret"]):
            # Store our Client Secret
            results["client_secret"] = \
                NotifySendPulse.unquote(results["qsd"]["secret"])

        elif results["targets"]:
            # Store our Client Secret
            results["client_secret"] = results["targets"].pop(0)

        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"].append(
                NotifySendPulse.unquote(results["qsd"]["to"]))

        # Handle Carbon Copy Addresses
        if "cc" in results["qsd"] and len(results["qsd"]["cc"]):
            results["cc"] = \
                NotifySendPulse.unquote(results["qsd"]["cc"])

        # Handle Blind Carbon Copy Addresses
        if "bcc" in results["qsd"] and len(results["qsd"]["bcc"]):
            results["bcc"] = \
                NotifySendPulse.unquote(results["qsd"]["bcc"])

        # Handle Blind Carbon Copy Addresses
        if "template" in results["qsd"] and len(results["qsd"]["template"]):
            results["template"] = \
                NotifySendPulse.unquote(results["qsd"]["template"])

        # Add any template substitutions
        results["template_data"] = results["qsd+"]

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from itertools import chain

try:
    import smpplib
    import smpplib.consts
    import smpplib.gsm

    # We're good to go!
    NOTIFY_SMPP_ENABLED = True

except ImportError:
    # cryptography is required in order for this package to work
    NOTIFY_SMPP_ENABLED = False


from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import is_phone_no, parse_phone_no
from .base import NotifyBase


class NotifySMPP(NotifyBase):
    """A wrapper for SMPP Notifications."""

    # Set our global enabled flag
    enabled = NOTIFY_SMPP_ENABLED

    requirements = {
        # Define our required packaging in order to work
        "packages_required": "smpplib"
    }

    # The default descriptive name associated with the Notification
    service_name = _("SMPP")

    # The services URL
    service_url = "https://smpp.org/"

    # The default protocol
    protocol = "smpp"

    # The default secure protocol
    secure_protocol = "smpps"

    # Default port setup
    default_port = 2775
    default_secure_port = 3550

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_SMPP"

    # A title can not be used for SMS Messages.  Setting this to zero will
    # cause any title (if defined) to get placed into the message body.
    title_maxlen = 0

    templates = (
        "{schema}://{user}:{password}@{host}/{from_phone}/{targets}",
        "{schema}://{user}:{password}@{host}:{port}/{from_phone}/{targets}",
    )

    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "user": {
                "name": _("Username"),
                "type": "string",
                "required": True,
            },
            "password": {
                "name": _("Password"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "host": {
                "name": _("Host"),
                "type": "string",
                "required": True,
            },
            "port": {
                "name": _("Port"),
                "type": "int",
                "min": 1,
                "max": 65535,
            },
            "from_phone": {
                "name": _("From Phone No"),
                "type": "string",
                "regex": (r"^[0-9\s)(+-]+$", "i"),
                "required": True,
                "map_to": "source",
            },
            "target_phone": {
                "name": _("Target Phone No"),
                "type": "string",
                "prefix": "+",
                "regex": (r"^[0-9\s)(+-]+$", "i"),
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
                "required": True,
            },
        },
    )

    def __init__(self, source=None, targets=None, **kwargs):
        """Initialize SMPP Object."""

        super().__init__(**kwargs)

        self.source = None

        if not (self.user and self.password):
            msg = "No SMPP user/pass combination was provided"
            self.logger.warning(msg)
            raise TypeError(msg)

        result = is_phone_no(source)
        if not result:
            msg = (
                f"The Account (From) Phone # specified ({source}) is invalid."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        # Tidy source
        self.source = result["full"]

        # Used for URL generation afterwards only
        self._invalid_targets = []

        # Parse our targets
        self.targets = []

        for target in parse_phone_no(targets, prefix=True):
            # Validate targets and drop bad ones:
            result = is_phone_no(target)
            if not result:
                self.logger.warning(
                    f"Dropped invalid phone # ({target}) specified.",
                )
                self._invalid_targets.append(target)
                continue

            # store valid phone number
            self.targets.append(result["full"])

    @property
    def url_identifier(self):
        """Returns all the identifiers that make this URL unique from another
        similar one.

        Targets or end points should never be identified here.
        """
        return (
            self.secure_protocol if self.secure else self.protocol,
            self.user,
            self.password,
            self.host,
            self.port,
            self.source,
        )

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        params = self.url_parameters(privacy=privacy, *args, **kwargs)

        return (
            "{schema}://{user}:{password}@{host}/{source}/{targets}/?{params}"
        ).format(
            schema=self.secure_protocol if self.secure else self.protocol,
            user=self.user,
            password=self.password,
            host=f"{self.host}:{self.port}" if self.port else self.host,
            source=self.source,
            targets="/".join([
                NotifySMPP.quote(t, safe="")
                for t in chain(self.targets, self._invalid_targets)
            ]),
            params=self.urlencode(params),
        )

    def __len__(self):
        """Returns the number of targets associated with this notification.

        Always return 1 at least
        """
        return len(self.targets) if self.targets else 1

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform SMPP Notification."""

        if not self.targets:
            # There were no targets to notify
            self.logger.warning("There were no SMPP targets to notify")
            return False

        # error tracking (used for function return)
        has_error = False

        port = (
            self.default_port if not self.secure else self.default_secure_port
        )

        client = smpplib.client.Client(
            self.host, port, allow_unknown_opt_params=True
        )
        try:
            client.connect()
            client.bind_transmitter(
                system_id=self.user, password=self.password
            )

        except smpplib.exceptions.ConnectionError as e:
            self.logger.warning(
                "Failed to establish connection to SMPP server"
                f" {self.host}: {e}"
            )
            return False

        for target in self.targets:
            parts, encoding, msg_type = smpplib.gsm.make_parts(body)

            # Always call throttle before any remote server i/o is made
            self.throttle()

            try:
                for payload in parts:
                    client.send_message(
                        source_addr_ton=smpplib.consts.SMPP_TON_INTL,
                        source_addr=self.source,
                        dest_addr_ton=smpplib.consts.SMPP_TON_INTL,
                        destination_addr=target,
                        short_message=payload,
                        data_coding=encoding,
                        esm_class=msg_type,
                        registered_delivery=True,
                    )
            except Exception as e:
                self.logger.warning(f"Failed to send SMPP notification: {e}")
                # Mark our failure
                has_error = True
                continue

            self.logger.info("Sent SMPP notification to %s", target)

        client.unbind()
        client.disconnect()
        return not has_error

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Support the 'from'  and 'source' variable so that we can support
        # targets this way too.
        # The 'from' makes it easier to use yaml configuration
        if "from" in results["qsd"] and len(results["qsd"]["from"]):
            results["source"] = NotifySMPP.unquote(results["qsd"]["from"])

            # hostname will also be a target in this case
            results["targets"] = [
                *NotifySMPP.parse_phone_no(results["host"]),
                *NotifySMPP.split_path(results["fullpath"]),
            ]

        else:
            # store our source
            results["source"] = NotifySMPP.unquote(results["host"])

            # store targets
            results["targets"] = NotifySMPP.split_path(results["fullpath"])

        # Support the 'to' variable so that we can support targets this way too
        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += NotifySMPP.parse_phone_no(
                results["qsd"]["to"]
            )

        results["targets"] = NotifySMPP.split_path(results["fullpath"])

        # Support the 'to' variable so that we can support targets this way too
        # 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += NotifySMPP.parse_phone_no(
                results["qsd"]["to"]
            )

        # store any additional payload extras defined
        results["payload"] = {
            NotifySMPP.unquote(x): NotifySMPP.unquote(y)
            for x, y in results["qsd:"].items()
        }

        # Add our GET parameters in the event the user wants to pass them
        results["params"] = {
            NotifySMPP.unquote(x): NotifySMPP.unquote(y)
            for x, y in results["qsd-"].items()
        }

        # Support the 'from' and 'source' variable so that we can support
        # targets this way too.
        # 'from' makes it easier to use yaml configuration
        if "from" in results["qsd"] and len(results["qsd"]["from"]):
            results["source"] = NotifySMPP.unquote(results["qsd"]["from"])
        elif results["targets"]:
            # from phone number is the first entry in the list otherwise
            results["source"] = results["targets"].pop(0)

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# See https://developer.twitter.com/en/docs/direct-messages/\
#           sending-and-receiving/api-reference/new-event.html
import contextlib
from copy import deepcopy
from datetime import datetime, timezone
from json import dumps, loads
import re

import requests
from requests_oauthlib import OAuth1

from ..attachment.base import AttachBase
from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..url import PrivacyMode
from ..utils.parse import parse_bool, parse_list, validate_regex
from .base import NotifyBase

IS_USER = re.compile(r"^\s*@?(?P<user>[A-Z0-9_]+)$", re.I)


class TwitterMessageMode:
    """Twitter Message Mode."""

    # DM (a Direct Message)
    DM = "dm"

    # A Public Tweet
    TWEET = "tweet"


# Define the types in a list for validation purposes
TWITTER_MESSAGE_MODES = (
    TwitterMessageMode.DM,
    TwitterMessageMode.TWEET,
)


class NotifyTwitter(NotifyBase):
    """A wrapper to Twitter Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Twitter"

    # The services URL
    service_url = "https://twitter.com/"

    # The default secure protocol is twitter.
    secure_protocol = ("x", "twitter", "tweet")

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_twitter"

    # Support attachments
    attachment_support = True

    # Do not set body_maxlen as it is set in a property value below
    # since the length varies depending if we are doing a direct message
    # or a tweet
    # body_maxlen = see below @propery defined

    # Twitter does have titles when creating a message
    title_maxlen = 0

    # Twitter API Reference To Acquire Someone's Twitter ID
    twitter_lookup = "https://api.twitter.com/1.1/users/lookup.json"

    # Twitter API Reference To Acquire Current Users Information
    twitter_whoami = (
        "https://api.twitter.com/1.1/account/verify_credentials.json"
    )

    # Twitter API Reference To Send A Private DM
    twitter_dm = "https://api.twitter.com/1.1/direct_messages/events/new.json"

    # Twitter API Reference To Send A Public Tweet
    twitter_tweet = "https://api.twitter.com/1.1/statuses/update.json"

    # it is documented on the site that the maximum images per tweet
    # is 4 (unless it's a GIF, then it's only 1)
    __tweet_non_gif_images_batch = 4

    # Twitter Media (Attachment) Upload Location
    twitter_media = "https://upload.twitter.com/1.1/media/upload.json"

    # Twitter is kind enough to return how many more requests we're allowed to
    # continue to make within it's header response as:
    # X-Rate-Limit-Reset: The epoc time (in seconds) we can expect our
    #                    rate-limit to be reset.
    # X-Rate-Limit-Remaining: an integer identifying how many requests we're
    #                        still allow to make.
    request_rate_per_sec = 0

    # For Tracking Purposes
    ratelimit_reset = datetime.now(timezone.utc).replace(tzinfo=None)

    # Default to 1000; users can send up to 1000 DM's and 2400 tweets a day
    # This value only get's adjusted if the server sets it that way
    ratelimit_remaining = 1

    templates = (
        "{schema}://{ckey}/{csecret}/{akey}/{asecret}",
        "{schema}://{ckey}/{csecret}/{akey}/{asecret}/{targets}",
    )

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "ckey": {
                "name": _("Consumer Key"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "csecret": {
                "name": _("Consumer Secret"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "akey": {
                "name": _("Access Key"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "asecret": {
                "name": _("Access Secret"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "target_user": {
                "name": _("Target User"),
                "type": "string",
                "prefix": "@",
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "mode": {
                "name": _("Message Mode"),
                "type": "choice:string",
                "values": TWITTER_MESSAGE_MODES,
                "default": TwitterMessageMode.DM,
            },
            "cache": {
                "name": _("Cache Results"),
                "type": "bool",
                "default": True,
            },
            "to": {
                "alias_of": "targets",
            },
            "batch": {
                "name": _("Batch Mode"),
                "type": "bool",
                "default": True,
            },
        },
    )

    def __init__(
        self,
        ckey,
        csecret,
        akey,
        asecret,
        targets=None,
        mode=None,
        cache=True,
        batch=True,
        **kwargs,
    ):
        """Initialize Twitter Object."""
        super().__init__(**kwargs)

        self.ckey = validate_regex(ckey)
        if not self.ckey:
            msg = "An invalid Twitter Consumer Key was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        self.csecret = validate_regex(csecret)
        if not self.csecret:
            msg = "An invalid Twitter Consumer Secret was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        self.akey = validate_regex(akey)
        if not self.akey:
            msg = "An invalid Twitter Access Key was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        self.asecret = validate_regex(asecret)
        if not self.asecret:
            msg = "An invalid Access Secret was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        # Store our webhook mode
        self.mode = (
            self.template_args["mode"]["default"]
            if not isinstance(mode, str)
            else mode.lower()
        )

        if mode and isinstance(mode, str):
            self.mode = next(
                (a for a in TWITTER_MESSAGE_MODES if a.startswith(mode)), None
            )
            if self.mode not in TWITTER_MESSAGE_MODES:
                msg = (
                    f"The Twitter message mode specified ({mode}) is invalid."
                )
                self.logger.warning(msg)
                raise TypeError(msg)
        else:
            self.mode = self.template_args["mode"]["default"]

        # Set Cache Flag
        self.cache = cache

        # Prepare Image Batch Mode Flag
        self.batch = batch

        # Track any errors
        has_error = False

        # Identify our targets
        self.targets = []
        for target in parse_list(targets):
            match = IS_USER.match(target)
            if match and match.group("user"):
                self.targets.append(match.group("user"))
                continue

            has_error = True
            self.logger.warning(
                f"Dropped invalid Twitter user ({target}) specified.",
            )

        if has_error and not self.targets:
            # We have specified that we want to notify one or more individual
            # and we failed to load any of them.  Since it's also valid to
            # notify no one at all (which means we notify ourselves), it's
            # important we don't switch from the users original intentions
            self.targets = None

        # Initialize our cache values
        self._whoami_cache = None
        self._user_cache = {}

        return

    def send(
        self,
        body,
        title="",
        notify_type=NotifyType.INFO,
        attach=None,
        **kwargs,
    ):
        """Perform Twitter Notification."""

        if self.targets is None:
            self.logger.warning("No valid Twitter targets to notify.")
            return False

        # Build a list of our attachments
        attachments = []

        if attach and self.attachment_support:
            # We need to upload our payload first so that we can source it
            # in remaining messages
            for no, attachment in enumerate(attach, start=1):

                # Perform some simple error checking
                if not attachment:
                    # We could not access the attachment
                    self.logger.error(
                        "Could not access attachment "
                        f"'{attachment.url(privacy=True)}."
                    )
                    return False

                if not re.match(r"^image/.*", attachment.mimetype, re.I):
                    # Only support images at this time
                    self.logger.warning(
                        "Ignoring unsupported Twitter attachment "
                        f"{attachment.url(privacy=True)}."
                    )
                    continue

                self.logger.debug(
                    "Preparing Twitter attachment "
                    f"{attachment.url(privacy=True)}"
                )

                # Upload our image and get our id associated with it
                # see: https://developer.twitter.com/en/docs/twitter-api/v1/\
                #         media/upload-media/api-reference/post-media-upload
                postokay, response = self._fetch(
                    self.twitter_media,
                    payload=attachment,
                )

                if not postokay:
                    # We can't post our attachment
                    return False

                # Prepare our filename
                filename = (
                    attachment.name if attachment.name else f"file{no:03}.dat"
                )

                if not (
                    isinstance(response, dict) and response.get("media_id")
                ):
                    self.logger.debug(
                        "Could not attach the file to Twitter: %s (mime=%s)",
                        filename,
                        attachment.mimetype,
                    )
                    continue

                # If we get here, our output will look something like this:
                # {
                #   "media_id": 710511363345354753,
                #   "media_id_string": "710511363345354753",
                #   "media_key": "3_710511363345354753",
                #   "size": 11065,
                #   "expires_after_secs": 86400,
                #   "image": {
                #     "image_type": "image/jpeg",
                #     "w": 800,
                #     "h": 320
                #   }
                # }

                response.update({
                    # Update our response to additionally include the
                    # attachment details
                    "file_name": filename,
                    "file_mime": attachment.mimetype,
                    "file_path": attachment.path,
                })

                # Save our pre-prepared payload for attachment posting
                attachments.append(response)

        # - calls _send_tweet if the mode is set so
        # - calls _send_dm (direct message) otherwise
        return getattr(self, f"_send_{self.mode}")(
            body=body,
            title=title,
            notify_type=notify_type,
            attachments=attachments,
            **kwargs,
        )

    def _send_tweet(
        self,
        body,
        title="",
        notify_type=NotifyType.INFO,
        attachments=None,
        **kwargs,
    ):
        """Twitter Public Tweet."""

        # Error Tracking
        has_error = False

        payload = {
            "status": body,
        }

        payloads = []
        if not attachments:
            payloads.append(payload)

        else:
            # Group our images if batch is set to do so
            batch_size = (
                1 if not self.batch else self.__tweet_non_gif_images_batch
            )

            # Track our batch control in our message generation
            batches = []
            batch = []
            for attachment in attachments:
                batch.append(str(attachment["media_id"]))

                # Twitter supports batching images together.  This allows
                # the batching of multiple images together.  Twitter also
                # makes it clear that you can't batch `gif` files; they need
                # to be separate.  So the below preserves the ordering that
                # a user passed their attachments in.  if 4-non-gif images
                # are passed, they are all part of a single message.
                #
                # however, if they pass in image, gif, image, gif.  The
                # gif's inbetween break apart the batches so this would
                # produce 4 separate tweets.
                #
                # If you passed in, image, image, gif, image. <- This would
                # produce 3 images (as the first 2 images could be lumped
                # together as a batch)
                if (
                    not re.match(
                        r"^image/(png|jpe?g)", attachment["file_mime"], re.I
                    )
                    or len(batch) >= batch_size
                ):
                    batches.append(",".join(batch))
                    batch = []

            if batch:
                batches.append(",".join(batch))

            for no, media_ids in enumerate(batches):
                _payload = deepcopy(payload)
                _payload["media_ids"] = media_ids

                if no or not body:
                    # strip text and replace it with the image representation
                    _payload["status"] = f"{no + 1:02d}/{len(batches):02d}"
                payloads.append(_payload)

        for no, payload in enumerate(payloads, start=1):
            # Send Tweet
            postokay, response = self._fetch(
                self.twitter_tweet,
                payload=payload,
                json=False,
            )

            if not postokay:
                # Track our error
                has_error = True

                errors = []
                with contextlib.suppress(KeyError, TypeError):
                    errors = [
                        "Error Code {}: {}".format(
                            e.get("code", "unk"), e.get("message")
                        )
                        for e in response["errors"]
                    ]

                for error in errors:
                    self.logger.debug(
                        "Tweet [%.2d/%.2d] Details: %s",
                        no,
                        len(payloads),
                        error,
                    )
                continue

            try:
                url = "https://twitter.com/{}/status/{}".format(
                    response["user"]["screen_name"], response["id_str"]
                )

            except (KeyError, TypeError):
                url = "unknown"

            self.logger.debug(
                "Tweet [%.2d/%.2d] Details: %s", no, len(payloads), url
            )

            self.logger.info(
                "Sent [%.2d/%.2d] Twitter notification as public tweet.",
                no,
                len(payloads),
            )

        return not has_error

    def _send_dm(
        self,
        body,
        title="",
        notify_type=NotifyType.INFO,
        attachments=None,
        **kwargs,
    ):
        """Twitter Direct Message."""

        # Error Tracking
        has_error = False

        payload = {
            "event": {
                "type": "message_create",
                "message_create": {
                    "target": {
                        # This gets assigned
                        "recipient_id": None,
                    },
                    "message_data": {
                        "text": body,
                    },
                },
            }
        }

        # Lookup our users (otherwise we look up ourselves)
        targets = (
            self._whoami(lazy=self.cache)
            if not len(self.targets)
            else self._user_lookup(self.targets, lazy=self.cache)
        )

        if not targets:
            # We failed to lookup any users
            self.logger.warning(
                "Failed to acquire user(s) to Direct Message via Twitter"
            )
            return False

        payloads = []
        if not attachments:
            payloads.append(payload)

        else:
            for no, attachment in enumerate(attachments):
                _payload = deepcopy(payload)
                _data = _payload["event"]["message_create"]["message_data"]
                _data["attachment"] = {
                    "type": "media",
                    "media": {"id": attachment["media_id"]},
                    "additional_owners": ",".join(
                        [str(x) for x in targets.values()]
                    ),
                }
                if no or not body:
                    # strip text and replace it with the image representation
                    _data["text"] = f"{no + 1:02d}/{len(attachments):02d}"
                payloads.append(_payload)

        for no, payload in enumerate(payloads, start=1):
            for screen_name, user_id in targets.items():
                # Assign our user
                target = payload["event"]["message_create"]["target"]
                target["recipient_id"] = user_id

                # Send Twitter DM
                postokay, _response = self._fetch(
                    self.twitter_dm,
                    payload=payload,
                )

                if not postokay:
                    # Track our error
                    has_error = True
                    continue

                self.logger.info(
                    f"Sent [{no:02d}/{len(payloads):02d}] "
                    f"Twitter DM notification to @{screen_name}."
                )

        return not has_error

    def _whoami(self, lazy=True):
        """Looks details of current authenticated user."""

        if lazy and self._whoami_cache is not None:
            # Use cached response
            return self._whoami_cache

        # Contains a mapping of screen_name to id
        results = {}

        # Send Twitter DM
        postokay, response = self._fetch(
            self.twitter_whoami,
            method="GET",
            json=False,
        )

        if postokay:
            try:
                results[response["screen_name"]] = response["id"]
                self._whoami_cache = {
                    response["screen_name"]: response["id"],
                }

                self._user_cache.update(results)

            except (TypeError, KeyError):
                pass

        return results

    def _user_lookup(self, screen_name, lazy=True):
        """Looks up a screen name and returns the user id.

        the screen_name can be a list/set/tuple as well
        """

        # Contains a mapping of screen_name to id
        results = {}

        # Build a unique set of names
        names = parse_list(screen_name)

        if lazy and self._user_cache:
            # Use cached response
            results = {k: v for k, v in self._user_cache.items() if k in names}

            # limit our names if they already exist in our cache
            names = [name for name in names if name not in results]

        if not len(names):
            # They're is nothing further to do
            return results

        # Twitters API documents that it can lookup to 100
        # results at a time.
        # https://developer.twitter.com/en/docs/accounts-and-users/\
        #     follow-search-get-users/api-reference/get-users-lookup
        for i in range(0, len(names), 100):
            # Look up our names by their screen_name
            postokay, response = self._fetch(
                self.twitter_lookup,
                payload={
                    "screen_name": names[i : i + 100],
                },
                json=False,
            )

            if not postokay or not isinstance(response, list):
                # Track our error
                continue

            # Update our user index
            for entry in response:
                with contextlib.suppress(TypeError, KeyError):
                    results[entry["screen_name"]] = entry["id"]

        # Cache our response for future use; this saves on un-nessisary extra
        # hits against the Twitter API when we already know the answer
        self._user_cache.update(results)

        return results

    def _fetch(self, url, payload=None, method="POST", json=True):
        """Wrapper to Twitter API requests object."""

        headers = {
            "User-Agent": self.app_id,
        }

        data = None
        files = None

        # Open our attachment path if required:
        if isinstance(payload, AttachBase):
            # prepare payload
            files = {
                "media": (
                    payload.name,
                    # file handle is safely closed in `finally`; inline open is
                    # intentional
                    open(payload.path, "rb"),  # noqa: SIM115
                ),
            }

        elif json:
            headers["Content-Type"] = "application/json"
            data = dumps(payload)

        else:
            data = payload

        auth = OAuth1(
            self.ckey,
            client_secret=self.csecret,
            resource_owner_key=self.akey,
            resource_owner_secret=self.asecret,
        )

        # Some Debug Logging
        self.logger.debug(
            f"Twitter {method} URL: {url} "
            f"(cert_verify={self.verify_certificate})"
        )
        self.logger.debug(f"Twitter Payload: {payload!s}")

        # By default set wait to None
        wait = None

        if self.ratelimit_remaining == 0:
            # Determine how long we should wait for or if we should wait at
            # all. This isn't fool-proof because we can't be sure the client
            # time (calling this script) is completely synced up with the
            # Twitter server.  One would hope we're on NTP and our clocks are
            # the same allowing this to role smoothly:

            now = datetime.now(timezone.utc).replace(tzinfo=None)
            if now < self.ratelimit_reset:
                # We need to throttle for the difference in seconds
                # We add 0.5 seconds to the end just to allow a grace
                # period.
                wait = (self.ratelimit_reset - now).total_seconds() + 0.5

        # Default content response object
        content = {}

        # Always call throttle before any remote server i/o is made;
        self.throttle(wait=wait)

        # acquire our request mode
        fn = requests.post if method == "POST" else requests.get
        try:
            r = fn(
                url,
                data=data,
                files=files,
                headers=headers,
                auth=auth,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )

            try:
                content = loads(r.content)

            except (AttributeError, TypeError, ValueError):
                # ValueError = r.content is Unparsable
                # TypeError = r.content is None
                # AttributeError = r is None
                content = {}

            if r.status_code != requests.codes.ok:
                # We had a problem
                status_str = NotifyTwitter.http_response_code_lookup(
                    r.status_code
                )

                self.logger.warning(
                    "Failed to send Twitter {} to {}: {}error={}.".format(
                        method, url, ", " if status_str else "", r.status_code
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                # Mark our failure
                return (False, content)

            try:
                # Capture rate limiting if possible
                self.ratelimit_remaining = int(
                    r.headers.get("x-rate-limit-remaining")
                )
                self.ratelimit_reset = datetime.fromtimestamp(
                    int(r.headers.get("x-rate-limit-reset")), timezone.utc
                ).replace(tzinfo=None)

            except (TypeError, ValueError):
                # This is returned if we could not retrieve this information
                # gracefully accept this state and move on
                pass

        except requests.RequestException as e:
            self.logger.warning(
                f"Exception received when sending Twitter {method} to {url}: "
            )
            self.logger.debug(f"Socket Exception: {e!s}")

            # Mark our failure
            return (False, content)

        except OSError as e:
            self.logger.warning(
                "An I/O error occurred while handling {}.".format(
                    payload.name
                    if isinstance(payload, AttachBase)
                    else payload
                )
            )
            self.logger.debug(f"I/O Exception: {e!s}")
            return (False, content)

        finally:
            # Close our file (if it's open) stored in the second element
            # of our files tuple (index 1)
            if files:
                files["media"][1].close()

        return (True, content)

    @property
    def body_maxlen(self):
        """The maximum allowable characters allowed in the body per message
        This is used during a Private DM Message Size (not Public Tweets which
        are limited to 280 characters)"""
        return 10000 if self.mode == TwitterMessageMode.DM else 280

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (
            self.secure_protocol[0],
            self.ckey,
            self.csecret,
            self.akey,
            self.asecret,
        )

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "mode": self.mode,
            "batch": "yes" if self.batch else "no",
            "cache": "yes" if self.cache else "no",
        }

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        return (
            "{schema}://{ckey}/{csecret}/{akey}/{asecret}"
            "/{targets}?{params}".format(
                schema=self.secure_protocol[0],
                ckey=self.pprint(self.ckey, privacy, safe=""),
                csecret=self.pprint(
                    self.csecret, privacy, mode=PrivacyMode.Secret, safe=""
                ),
                akey=self.pprint(self.akey, privacy, safe=""),
                asecret=self.pprint(
                    self.asecret, privacy, mode=PrivacyMode.Secret, safe=""
                ),
                targets=(
                    "/".join([
                        NotifyTwitter.quote(f"@{target}", safe="@")
                        for target in self.targets
                    ])
                    if self.targets
                    else ""
                ),
                params=NotifyTwitter.urlencode(params),
            )
        )

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        targets = len(self.targets)
        return targets if targets > 0 else 1

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Acquire remaining tokens
        tokens = NotifyTwitter.split_path(results["fullpath"])

        # The consumer token is stored in the hostname
        results["ckey"] = NotifyTwitter.unquote(results["host"])

        #
        # Now fetch the remaining tokens
        #

        # Consumer Secret
        results["csecret"] = tokens.pop(0) if tokens else None
        # Access Token Key
        results["akey"] = tokens.pop(0) if tokens else None
        # Access Token Secret
        results["asecret"] = tokens.pop(0) if tokens else None

        # The defined twitter mode
        if "mode" in results["qsd"] and len(results["qsd"]["mode"]):
            results["mode"] = NotifyTwitter.unquote(results["qsd"]["mode"])

        elif results["schema"].startswith("tweet"):
            results["mode"] = TwitterMessageMode.TWEET

        results["targets"] = []

        # if a user has been defined, add it to the list of targets
        if results.get("user"):
            results["targets"].append(results.get("user"))

        # Store any remaining items as potential targets
        results["targets"].extend(tokens)

        # Get Cache Flag (reduces lookup hits)
        if "cache" in results["qsd"] and len(results["qsd"]["cache"]):
            results["cache"] = parse_bool(results["qsd"]["cache"], True)

        # Get Batch Mode Flag
        results["batch"] = parse_bool(
            results["qsd"].get(
                "batch", NotifyTwitter.template_args["batch"]["default"]
            )
        )

        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += NotifyTwitter.parse_list(
                results["qsd"]["to"]
            )

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# To use this plugin, you must have a Kumulos account set up. Add a client
# and link it with your phone using the phone app (using your Companion App
# option in the profile menu area):
#    Android: https://play.google.com/store/apps/\
#                     details?id=com.kumulos.companion
#    iOS: https://apps.apple.com/us/app/kumulos/id1463947782
#
# The API reference used to build this plugin was documented here:
#  https://docs.kumulos.com/messaging/api/#sending-in-app-messages
#
from json import dumps

import requests

from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import validate_regex
from .base import NotifyBase

# Extend HTTP Error Messages
KUMULOS_HTTP_ERROR_MAP = {
    401: "Unauthorized - Invalid API and/or Server Key.",
    422: "Unprocessable Entity - The request was unparsable.",
    400: "Bad Request - Targeted users do not exist or have unsubscribed.",
}


class NotifyKumulos(NotifyBase):
    """A wrapper for Kumulos Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Kumulos"

    # The services URL
    service_url = "https://kumulos.com/"

    # The default secure protocol
    secure_protocol = "kumulos"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_kumulos"

    # Kumulos uses the http protocol with JSON requests
    notify_url = "https://messages.kumulos.com/v2/notifications"

    # The maximum allowable characters allowed in the title per message
    title_maxlen = 64

    # The maximum allowable characters allowed in the body per message
    body_maxlen = 240

    # Define object templates
    templates = ("{schema}://{apikey}/{serverkey}",)

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "apikey": {
                "name": _("API Key"),
                "type": "string",
                "private": True,
                "required": True,
                # UUID4
                "regex": (
                    r"^[0-9a-f]{8}-[0-9a-f]{4}-4[0-9a-f]{3}-"
                    r"[89ab][0-9a-f]{3}-[0-9a-f]{12}$",
                    "i",
                ),
            },
            "serverkey": {
                "name": _("Server Key"),
                "type": "string",
                "private": True,
                "required": True,
                "regex": (r"^[A-Z0-9+]{36}$", "i"),
            },
        },
    )

    def __init__(self, apikey, serverkey, **kwargs):
        """Initialize Kumulos Object."""
        super().__init__(**kwargs)

        # API Key (associated with project)
        self.apikey = validate_regex(
            apikey, *self.template_tokens["apikey"]["regex"]
        )
        if not self.apikey:
            msg = f"An invalid Kumulos API Key ({apikey}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        # Server Key (associated with project)
        self.serverkey = validate_regex(
            serverkey, *self.template_tokens["serverkey"]["regex"]
        )
        if not self.serverkey:
            msg = f"An invalid Kumulos Server Key ({serverkey}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform Kumulos Notification."""

        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/json",
            "Accept": "application/json",
        }

        # prepare JSON Object
        payload = {
            "target": {
                "broadcast": True,
            },
            "content": {
                "title": title,
                "message": body,
            },
        }

        # Determine Authentication
        auth = (self.apikey, self.serverkey)

        self.logger.debug(
            "Kumulos POST URL:"
            f" {self.notify_url} (cert_verify={self.verify_certificate!r})"
        )
        self.logger.debug(f"Kumulos Payload: {payload!s}")

        # Always call throttle before any remote server i/o is made
        self.throttle()
        try:
            r = requests.post(
                self.notify_url,
                data=dumps(payload),
                headers=headers,
                auth=auth,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )
            if r.status_code != requests.codes.ok:
                # We had a problem
                status_str = NotifyKumulos.http_response_code_lookup(
                    r.status_code, KUMULOS_HTTP_ERROR_MAP
                )

                self.logger.warning(
                    "Failed to send Kumulos notification: "
                    "{}{}error={}.".format(
                        status_str, ", " if status_str else "", r.status_code
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                return False

            else:
                self.logger.info("Sent Kumulos notification.")

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred sending Kumulos notification."
            )
            self.logger.debug(f"Socket Exception: {e!s}")

            return False
        return True

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (self.secure_protocol, self.apikey, self.serverkey)

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Our URL parameters
        params = self.url_parameters(privacy=privacy, *args, **kwargs)

        return "{schema}://{apikey}/{serverkey}/?{params}".format(
            schema=self.secure_protocol,
            apikey=self.pprint(self.apikey, privacy, safe=""),
            serverkey=self.pprint(self.serverkey, privacy, safe=""),
            params=NotifyKumulos.urlencode(params),
        )

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # The first token is stored in the hostname
        results["apikey"] = NotifyKumulos.unquote(results["host"])

        # Now fetch the remaining tokens
        try:
            results["serverkey"] = NotifyKumulos.split_path(
                results["fullpath"]
            )[0]

        except IndexError:
            # no token
            results["serverkey"] = None

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from .base import CustomNotifyPlugin


def notify(on, name=None):
    """
    @notify decorator allows you to map functions you've defined to be loaded
    as a regular notify by Apprise.  You must identify a protocol that
    users will trigger your call by.

        @notify(on="foobar")
        def your_declaration(body, title, notify_type, meta, *args, **kwargs):
            ...

    You can optionally provide the name to associate with the plugin which
    is what calling functions via the API will receive.

        @notify(on="foobar", name="My Foobar Process")
        def your_action(body, title, notify_type, meta, *args, **kwargs):
            ...

    The meta variable is actually the processed URL contents found in
    configuration files that landed you in this function you wrote in
    the first place.  It's very easily tokenized already for you so
    that you can bend the notification logic to your hearts content.

        @notify(on="foobar", name="My Foobar Process")
        def your_action(body, title, notify_type, body_format, meta, attach,
                        *args, **kwargs):
            ...

    Arguments break down as follows:
      body:        The message body associated with the notification
      title:       The message title associated with the notification
      notify_type: The message type (info, success, warning, and failure)
      body_format: The format of the incoming notification body. This is
                   either text, html, or markdown.
      meta:        Combines the URL arguments specified on the `on` call
                   with the ones loaded from a users configuration. This
                   is a dictionary that presents itself like this:
                    {
                      'schema': 'http',
                      'url': 'http://hostname',
                      'host': 'hostname',

                      'user': 'john',
                      'password': 'doe',
                      'port': 80,
                      'path': '/',
                      'fullpath': '/test.php',
                      'query': 'test.php',

                      'qsd': {'key': 'value', 'key2': 'value2'},

                      'asset': <AppriseAsset>,
                      'tag': set(),
                    }

                    Meta entries are ONLY present if found.  A simple URL
                    such as foobar:// would only produce the following:
                    {
                      'schema': 'foobar',
                      'url': 'foobar://',

                      'asset': <AppriseAsset>,
                      'tag': set(),
                    }

      attach:      An array AppriseAttachment objects (if any were provided)

      body_format: Defaults to the expected format output; By default this
                   will be TEXT unless over-ridden in the Apprise URL


    If you don't intend on using all of the parameters, your @notify() call
    # can be greatly simplified to just:

        @notify(on="foobar", name="My Foobar Process")
        def your_action(body, title, *args, **kwargs)

    Always end your wrappers declaration with *args and **kwargs to be future
    proof with newer versions of Apprise.

    Your wrapper should return True if processed the send() function as you
    expected and return False if not. If nothing is returned, then this is
    treated as as success (True).

    """

    def wrapper(func):
        """Instantiate our custom (notification) plugin."""

        # Generate
        CustomNotifyPlugin.instantiate_plugin(
            url=on, send_func=func, name=name
        )

        return func

    return wrapper
